import enum
import gzip
import os
from typing import Union

from . import env


class SupportedOrganisms(enum.Enum):
    # NOTE: these could be enumerated from loading the `schema_definition.yaml` and scraping the 'organism_ontology_term_id' constraints
    HOMO_SAPIENS = "NCBITaxon:9606"
    MUS_MUSCULUS = "NCBITaxon:10090"
    SARS_COV_2 = "NCBITaxon:2697049"
    ERCC = "NCBITaxon:32630"
    DROSOPHILA_MELANOGASTER = "NCBITaxon:7227"
    DANIO_RERIO = "NCBITaxon:7955"
    CAENORHABDITIS_ELEGANS = "NCBITaxon:6239"
    MACACA_FASCICULARIS = "NCBITaxon:9541"
    ORYCTOLAGUS_CUNICULUS = "NCBITaxon:9986"
    CALLITHRIX_JACCHUS = "NCBITaxon:9483"
    GORILLA_GORILLA = "NCBITaxon:9595"
    MACACA_MULATTA = "NCBITaxon:9544"
    PAN_TROGLODYTES = "NCBITaxon:9598"
    SUS_SCROFA = "NCBITaxon:9823"
    MICROCEBUS_MURINUS = "NCBITaxon:30608"
    RATTUS_NORVEGICUS = "NCBITaxon:10116"


def get_organism_from_feature_id(
    feature_id: str,
) -> Union[SupportedOrganisms, None]:
    """
    Determines organism based on which gene file the feature id was in

    :param str feature_id: the feature id

    :rtype Union[ontology.SypportedOrganisms, None]
    :return: the organism the feature id is from
    """

    for organism in SupportedOrganisms:
        gene_checker = get_gene_checker(organism)
        if gene_checker.is_valid_id(feature_id):
            return organism

    return None


class GeneChecker:
    """Handles checking gene ids, retrieves symbols"""

    GENE_FILES = {
        SupportedOrganisms.HOMO_SAPIENS: os.path.join(env.GENCODE_DIR, "genes_homo_sapiens.csv.gz"),
        SupportedOrganisms.MUS_MUSCULUS: os.path.join(env.GENCODE_DIR, "genes_mus_musculus.csv.gz"),
        SupportedOrganisms.SARS_COV_2: os.path.join(env.GENCODE_DIR, "genes_sars_cov_2.csv.gz"),
        SupportedOrganisms.ERCC: os.path.join(env.GENCODE_DIR, "genes_ercc.csv.gz"),
        SupportedOrganisms.DROSOPHILA_MELANOGASTER: os.path.join(
            env.GENCODE_DIR, "genes_drosophila_melanogaster.csv.gz"
        ),
        SupportedOrganisms.DANIO_RERIO: os.path.join(env.GENCODE_DIR, "genes_danio_rerio.csv.gz"),
        SupportedOrganisms.CAENORHABDITIS_ELEGANS: os.path.join(env.GENCODE_DIR, "genes_caenorhabditis_elegans.csv.gz"),
        SupportedOrganisms.MACACA_FASCICULARIS: os.path.join(env.GENCODE_DIR, "genes_macaca_fascicularis.csv.gz"),
        SupportedOrganisms.ORYCTOLAGUS_CUNICULUS: os.path.join(env.GENCODE_DIR, "genes_oryctolagus_cuniculus.csv.gz"),
        SupportedOrganisms.CALLITHRIX_JACCHUS: os.path.join(env.GENCODE_DIR, "genes_callithrix_jacchus.csv.gz"),
        SupportedOrganisms.GORILLA_GORILLA: os.path.join(env.GENCODE_DIR, "genes_gorilla_gorilla.csv.gz"),
        SupportedOrganisms.MACACA_MULATTA: os.path.join(env.GENCODE_DIR, "genes_macaca_mulatta.csv.gz"),
        SupportedOrganisms.PAN_TROGLODYTES: os.path.join(env.GENCODE_DIR, "genes_pan_troglodytes.csv.gz"),
        SupportedOrganisms.SUS_SCROFA: os.path.join(env.GENCODE_DIR, "genes_sus_scrofa.csv.gz"),
        SupportedOrganisms.MICROCEBUS_MURINUS: os.path.join(env.GENCODE_DIR, "genes_microcebus_murinus.csv.gz"),
        SupportedOrganisms.RATTUS_NORVEGICUS: os.path.join(env.GENCODE_DIR, "genes_rattus_norvegicus.csv.gz"),
    }

    def __init__(self, species: SupportedOrganisms):
        """
        :param enum.Enum.SupportedSpecies species: item from SupportedOrganisms
        """

        if species not in self.GENE_FILES:
            raise ValueError(f"{species} not supported.")

        self.species = species
        self.gene_dict = {}
        with gzip.open(self.GENE_FILES[species], "rt") as genes:
            for gene in genes:
                gene = gene.rstrip().split(",")  # type: ignore
                gene_id = gene[0]
                gene_label = gene[1]
                gene_length = int(gene[3])
                gene_type = gene[4]

                self.gene_dict[gene_id] = (gene_label, gene_length, gene_type)

    def is_valid_id(self, gene_id: str) -> bool:
        """
        Checks for validity of gene id

        :param str gene_id: ENSEMBL gene id

        :rtype bool
        :return True if the gene_id is a valid ENSEMBL id, False otherwise
        """

        return gene_id in self.gene_dict

    def get_symbol(self, gene_id: str) -> str:
        """
        Gets symbol associated to the ENSEBML id

        :param str gene_id: ENSEMBL gene id

        :rtype str
        :return A gene symbol
        """

        if self.is_valid_id(gene_id):
            return self.gene_dict[gene_id][0]
        else:
            raise ValueError(f"The id '{gene_id}' is not a valid ENSEMBL id for '{self.species}'")

    def get_length(self, gene_id: str) -> int:
        """
        Gets feature length associated to the ENSEBML id

        :param str gene_id: ENSEMBL gene id

        :rtype int
        :return A gene length
        """

        if self.is_valid_id(gene_id):
            return self.gene_dict[gene_id][1]
        else:
            raise ValueError(f"The id '{gene_id}' is not a valid ENSEMBL id for '{self.species}'")

    def get_type(self, gene_id: str) -> str:
        """
        Gets feature type associated to the ENSEBML id

        :param str gene_id: ENSEMBL gene id

        :rtype str
        :return A feature type
        """

        if self.is_valid_id(gene_id):
            return self.gene_dict[gene_id][2]
        else:
            raise ValueError(f"The id '{gene_id}' is not a valid ENSEMBL id for '{self.species}'")


# cache the gene checkers
_gene_checkers = {}


def get_gene_checker(species: SupportedOrganisms) -> GeneChecker:
    # Values will be instances of gencode.GeneChecker,
    # keys will be one of gencode.SupportedOrganisms
    if species not in _gene_checkers:
        _gene_checkers[species] = GeneChecker(species)
    return _gene_checkers[species]



from tempfile import TemporaryDirectory
from unittest.mock import patch

import anndata
from cellxgene_schema.migrate import migrate
from fixtures.examples_validate import adata_with_labels_unmigrated


class TestMigrate:
    def test_migrate(self):
        test_ONTOLOGY_TERM_MAPS = {
            "assay": {
                "assay:1": "assay:2",
            },
            "cell_type": {
                "cell_type:1": "cell_type:2",
            },
            "development_stage": {
                "development_stage:1": "development_stage:2",
            },
            "disease": {
                "disease:1": "disease:2",
            },
            "organism": {
                "organism:1": "organism:2",
            },
            "self_reported_ethnicity": {
                "sre:1": "sre:2",
            },
            "sex": {"sex:1": "sex:2"},
            "tissue": {"tissue:1": "tissue:2"},
        }
        with TemporaryDirectory() as tmp, patch("cellxgene_schema.migrate.DEPRECATED_FEATURE_IDS", ["DUMMY"]), patch(
            "cellxgene_schema.migrate.ONTOLOGY_TERM_MAPS", test_ONTOLOGY_TERM_MAPS
        ), patch("cellxgene_schema.migrate.GENCODE_MAPPER", {"ENSSASG00005000004": "ENSSASG00005000004_NEW"}):
            result_h5ad = tmp + "result.h5ad"
            test_h5ad = tmp + "test.h5ad"
            adata_with_labels_unmigrated.copy().write_h5ad(test_h5ad, compression="gzip")

            # Verify regular adata is what we expect before migration
            assert any(adata_with_labels_unmigrated.var.index.isin(["DUMMY"]))
            assert any(adata_with_labels_unmigrated.var.index.isin(["ENSSASG00005000004"]))
            assert not any(adata_with_labels_unmigrated.var.index.isin(["ENSSASG00005000004_NEW"]))
            assert adata_with_labels_unmigrated.X.shape == (2, 2)

            # Verify raw adata is what we expect before migration
            adata_raw_with_labels_unmigrated = anndata.AnnData(
                adata_with_labels_unmigrated.raw.X,
                var=adata_with_labels_unmigrated.raw.var,
                obs=adata_with_labels_unmigrated.obs,
            )
            assert any(adata_raw_with_labels_unmigrated.var.index.isin(["DUMMY"]))
            assert any(adata_raw_with_labels_unmigrated.var.index.isin(["ENSSASG00005000004"]))
            assert not any(adata_raw_with_labels_unmigrated.var.index.isin(["ENSSASG00005000004_NEW"]))
            assert adata_raw_with_labels_unmigrated.X.shape == (2, 2)

            migrate(
                input_file=test_h5ad,
                output_file=result_h5ad,
                collection_id="",
                dataset_id="",
            )

            # Verify regular adata is what we expect after migration
            adata = anndata.read_h5ad(result_h5ad)
            assert not any(adata.var.index.isin(["DUMMY"]))
            assert not any(adata.var.index.isin(["ENSSASG00005000004"]))
            assert any(adata.var.index.isin(["ENSSASG00005000004_NEW"]))
            assert adata.X.shape == (2, 1)

            # Verify raw adata is what we expect after migration
            raw_adata = anndata.AnnData(adata.raw.X, var=adata.raw.var, obs=adata.obs)
            assert not any(raw_adata.var.index.isin(["DUMMY"]))
            assert not any(raw_adata.var.index.isin(["ENSSASG00005000004"]))
            assert any(raw_adata.var.index.isin(["ENSSASG00005000004_NEW"]))
            assert raw_adata.X.shape == (2, 1)



import logging

import anndata as ad
from cellxgene_ontology_guide.ontology_parser import OntologyParser

logger = logging.getLogger(__name__)


def map_species(input_file, output_file):
    ontology_parser = OntologyParser("5.3.0")
    adata = ad.read_h5ad(input_file, backed="r")
    map_columns = [
        ("organism_cell_type_ontology_term_id", "cell_type_ontology_term_id", "CL"),
        ("organism_tissue_ontology_term_id", "tissue_ontology_term_id", "UBERON"),
    ]
    for map_from, map_to, ontology in map_columns:
        logger.info(f"Mapping {ontology} terms to {map_to}")
        if not hasattr(adata.obs, map_to):
            logger.info(f"No existing {map_to} column detected, creating one and defaulting to empty strings.")
            adata.obs[map_to] = ""
        for category, group in adata.obs.groupby(map_from):
            try:
                closest_terms = ontology_parser.get_closest_bridge_term_ids(category, ontology)
            except Exception:
                logger.error(f"{category} in {map_from} is not a valid ontology term ID, cannot find {ontology} map.")
                continue
            if len(closest_terms) < 1:
                logger.warning(f"{category} has no closest match for {ontology} in its ancestry tree.")
            elif len(closest_terms) == 1:
                adata.obs.loc[group.index, map_to] = closest_terms[0]
                logger.info(f"Setting {category} {map_to} rows to single closest match: {closest_terms[0]}")
            else:
                logger.warning(
                    f"{category} has multiple closest matches for {ontology} in its ancestry tree: "
                    f"{closest_terms}. Pick one based on biological context and assign manually."
                )
        adata.obs[map_to] = adata.obs[map_to].astype("category")

    logger.info(f"Mappings complete, writing to {output_file}")
    adata.write_h5ad(output_file, compression="gzip")



import pandas as pd
import pytest
from anndata import AnnData
from cellxgene_schema.utils import (
    get_hash_digest_column,
    map_ontology_term,
    read_h5ad,
    remap_deprecated_features,
    remove_deprecated_features,
    replace_ontology_term,
)
from fixtures.examples_validate import adata, adata_non_raw, h5ad_valid


@pytest.fixture
def adata_with_raw():
    return adata.copy()


@pytest.fixture
def adata_without_raw():
    return adata_non_raw.copy()


@pytest.fixture
def deprecated_features():
    return ["ERCC-00002", "ENSG00000127603"]


@pytest.fixture
def remapped_features():
    return {"ENSMUSG00000059552": "ENSMUSG00000059552_NEW"}


@pytest.fixture
def deprecated_term_map_with_replacement_match():
    return {"EFO:0009899": "EFO:0000001"}


@pytest.fixture
def deprecated_term_map_no_replacement_match():
    return {"EFO:0000002": "EFO:0000003"}


def test_remove_deprecated_features__with_raw(adata_with_raw, deprecated_features):
    # Call the function under test
    result = remove_deprecated_features(adata=adata_with_raw, deprecated=deprecated_features)

    # Check if the deprecated features are removed
    assert result.var_names.tolist() == [
        "ENSMUSG00000059552",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]
    assert result.raw.var_names.tolist() == [
        "ENSMUSG00000059552",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]


def test_remove_deprecated_features__without_raw(adata_without_raw, deprecated_features):
    # Verify existing fixtures don't contain the deprecated features
    assert adata_without_raw.var_names.tolist() == [
        "ERCC-00002",
        "ENSG00000127603",
        "ENSMUSG00000059552",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]

    # Call the function under test
    result = remove_deprecated_features(adata=adata_without_raw, deprecated=deprecated_features)

    # Check if the deprecated features are removed
    assert result.var_names.tolist() == [
        "ENSMUSG00000059552",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]
    assert result.raw is None


def test_remap_deprecated_features__with_raw(adata_with_raw, remapped_features):
    # Verify existing fixtures don't contain the deprecated features
    assert adata_with_raw.var_names.tolist() == [
        "ERCC-00002",
        "ENSG00000127603",
        "ENSMUSG00000059552",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]
    assert adata_with_raw.raw.var_names.tolist() == [
        "ERCC-00002",
        "ENSG00000127603",
        "ENSMUSG00000059552",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]

    # Call the function under test
    result = remap_deprecated_features(adata=adata_with_raw, remapped_features=remapped_features)

    # Check if the deprecated features are replaced
    assert result.var_names.tolist() == [
        "ERCC-00002",
        "ENSG00000127603",
        "ENSMUSG00000059552_NEW",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]
    assert result.raw.var_names.tolist() == [
        "ERCC-00002",
        "ENSG00000127603",
        "ENSMUSG00000059552_NEW",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]


def test_remap_deprecated_features__without_raw(adata_without_raw, remapped_features):
    # Call the function under test
    result = remap_deprecated_features(adata=adata_without_raw, remapped_features=remapped_features)

    # Check if the deprecated features are replaced
    assert result.var_names.tolist() == [
        "ERCC-00002",
        "ENSG00000127603",
        "ENSMUSG00000059552_NEW",
        "ENSSASG00005000004",
        "FBtr0472816_df_nrg",
        "ENSDARG00000009657",
        "WBGene00000003",
    ]
    assert result.raw is None


def test_replace_ontology_term__with_replacement(adata_with_raw, deprecated_term_map_with_replacement_match):
    replace_ontology_term(adata_with_raw.obs, "assay", deprecated_term_map_with_replacement_match)

    expected = ["EFO:0000001", "EFO:0008992"]
    actual = adata_with_raw.obs["assay_ontology_term_id"].dtype.categories
    assert sorted(actual) == expected


def test_replace_ontology_term__no_replacement(adata_with_raw, deprecated_term_map_no_replacement_match):
    replace_ontology_term(adata_with_raw.obs, "assay", deprecated_term_map_no_replacement_match)
    expected = ["EFO:0008992", "EFO:0009899"]
    actual = adata_with_raw.obs["assay_ontology_term_id"].dtype.categories
    print(actual)
    assert all(a == b for a, b in zip(actual, expected))


def test_map_ontology_term__(adata_without_raw):
    update_map = {"donor_1": "CL:0000001", "donor_2": "CL:0000002"}
    map_ontology_term(adata_without_raw.obs, "cell_type", "donor_id", update_map)
    expected = ["CL:0000001", "CL:0000002"]
    actual = adata_without_raw.obs["cell_type_ontology_term_id"].dtype.categories
    assert all(a == b for a, b in zip(actual, expected))
    donor_1_rows = adata_without_raw.obs.loc[adata_without_raw.obs["donor_id"] == "donor_1"]
    assert all(a == "CL:0000001" for a in donor_1_rows["cell_type_ontology_term_id"])
    donor_2_rows = adata_without_raw.obs.loc[adata_without_raw.obs["donor_id"] == "donor_2"]
    assert all(a == "CL:0000002" for a in donor_2_rows["cell_type_ontology_term_id"])


class TestGetHashDigestColumn:
    def test_get_hash_digest_column(self, adata_with_raw):
        hash_digest_column = get_hash_digest_column(adata_with_raw.obs)
        expected_column = pd.Series(["ab6yl9v%fZ", "f-dZLjjiRl"], index=["X", "Y"])
        pd.testing.assert_series_equal(hash_digest_column, expected_column)


class TestReadH5AD:
    def test_read_h5ad(self):
        h5ad_path = h5ad_valid
        adata = read_h5ad(h5ad_path)
        assert isinstance(adata, AnnData)



import hashlib
import os
import re
import tempfile
from typing import Union
from unittest import mock

import anndata
import numpy as np
import pandas as pd
import pytest
from cellxgene_ontology_guide.entities import Ontology
from cellxgene_schema.schema import get_schema_definition
from cellxgene_schema.validate import (
    ERROR_SUFFIX_SPARSE_FORMAT,
    ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE,
    ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_FORBIDDEN,
    ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_IN_TISSUE_0,
    ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_NOTNULL,
    ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_REQUIRED,
    SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE,
    SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM,
    Validator,
    validate,
)
from cellxgene_schema.write_labels import AnnDataLabelAppender
from dask.array import from_array
from fixtures.examples_validate import adata as adata_valid
from fixtures.examples_validate import (
    adata_minimal,
    adata_slide_seqv2,
    adata_spatial_is_single_false,
    adata_visium,
    adata_with_labels,
    good_obs,
    good_uns,
    good_uns_with_visium_spatial,
    h5ad_invalid,
    h5ad_valid,
    visium_library_id,
)
from numpy import ndarray, zeros
from scipy import sparse
from scipy.sparse import spmatrix

# Tests for internal functions of the Validator and LabelWriter classes.


@pytest.fixture(scope="class")
def schema_def():
    return get_schema_definition()


@pytest.fixture()
def validator_with_minimal_adata():
    validator = Validator()
    validator.adata = adata_minimal.copy()
    return validator


@pytest.fixture
def label_writer(valid_adata):
    return AnnDataLabelAppender(adata_valid)


@pytest.fixture
def valid_adata():
    return adata_valid.copy()


class TestFieldValidation:
    def test_schema_definition(self, schema_def):
        """
        Tests that the definition of schema is well-defined
        """
        assert isinstance(schema_def["components"], dict)
        assert isinstance(schema_def["components"]["obs"], dict)
        assert isinstance(schema_def["components"]["obs"]["columns"], dict)

        # Check that any columns in obs that are "curie" have "curie_constraints" and "ontologies" under the constraints
        for i in schema_def["components"]["obs"]["columns"]:
            assert "type" in schema_def["components"]["obs"]["columns"][i]
            if schema_def["components"]["obs"]["columns"][i]["type"] == "curie":
                if "curie_constraints" in schema_def["components"]["obs"]["columns"][i]:
                    assert isinstance(
                        schema_def["components"]["obs"]["columns"][i]["curie_constraints"],
                        dict,
                    )
                    assert isinstance(
                        schema_def["components"]["obs"]["columns"][i]["curie_constraints"]["ontologies"],
                        list,
                    )

                    # Check that the allowed ontologies are in the ontology checker or 'NA' (special case)
                    for ontology_name in schema_def["components"]["obs"]["columns"][i]["curie_constraints"][
                        "ontologies"
                    ]:
                        if ontology_name != "NA":
                            assert getattr(Ontology, ontology_name, None) is not None
                else:
                    # if no curie_constraints in top-level for type curie, assert that 'dependencies' list exists
                    assert isinstance(
                        schema_def["components"]["obs"]["columns"][i]["dependencies"],
                        list,
                    )

    def test_validate_ontology_good(self, validator_with_minimal_adata, schema_def):
        validator = validator_with_minimal_adata
        column_name = "cell_type_ontology_term_id"
        curie_constraints = schema_def["components"]["obs"]["columns"][column_name]["curie_constraints"]
        validator._validate_curie("CL:0000066", column_name, curie_constraints)
        validator._validate_curie("CL:0000192", column_name, curie_constraints)
        assert not validator.errors

    def test_validate_ontology_wrong_ontology(self, validator_with_minimal_adata, schema_def):
        validator = validator_with_minimal_adata
        column_name = "cell_type_ontology_term_id"
        curie_constraints = schema_def["components"]["obs"]["columns"][column_name]["curie_constraints"]
        validator._validate_curie("EFO:0009899", column_name, curie_constraints)
        assert validator.errors

    def test_validate_ontology_wrong_term(self, validator_with_minimal_adata, schema_def):
        validator = validator_with_minimal_adata
        column_name = "cell_type_ontology_term_id"
        curie_constraints = schema_def["components"]["obs"]["columns"][column_name]["curie_constraints"]
        validator._validate_curie("NO_TERM2", column_name, curie_constraints)
        assert validator.errors


class TestAddLabelFunctions:
    def test_get_dictionary_mapping_feature_id(self, label_writer):
        # Good
        ids = [
            "ERCC-00002",
            "ENSG00000127603",
            "ENSMUSG00000059552",
            "ENSSASG00005000004",
            "FBtr0472816_df_nrg",
            "ENSDARG00000009657",
            "WBGene00000003",
        ]
        labels = ["ERCC-00002 (spike-in control)", "MACF1", "Trp53", "S", "FBtr0472816_df_nrg", "fgfr1op2", "aat-2"]
        expected_dict = dict(zip(ids, labels))
        assert label_writer._get_mapping_dict_feature_id(ids), expected_dict

        # Bad
        ids = ["NO_GENE"]
        with pytest.raises(ValueError):
            label_writer._get_mapping_dict_feature_id(ids)

    def test_get_dictionary_mapping_feature_reference(self, label_writer):
        # Good
        ids = [
            "ERCC-00002",
            "ENSG00000127603",
            "ENSMUSG00000059552",
            "ENSSASG00005000004",
            "FBtr0472816_df_nrg",
            "ENSDARG00000009657",
            "WBGene00000003",
        ]
        labels = [
            "NCBITaxon:32630",
            "NCBITaxon:9606",
            "NCBITaxon:10090",
            "NCBITaxon:2697049",
            "NCBITaxon:7227",
            "NCBITaxon:7955",
            "NCBITaxon:6239",
        ]
        expected_dict = dict(zip(ids, labels))
        assert label_writer._get_mapping_dict_feature_reference(ids) == expected_dict

        # Bad
        ids = ["NO_GENE"]
        with pytest.raises(ValueError):
            label_writer._get_mapping_dict_feature_id(ids)

    def test_get_dictionary_mapping_feature_length(self, label_writer):
        # Good
        ids = [
            "ERCC-00002",
            "ENSG00000127603",
            "ENSMUSG00000059552",
            "ENSSASG00005000004",
            "FBtr0472816_df_nrg",
            "ENSDARG00000009657",
            "WBGene00000003",
        ]
        # values derived from csv
        gene_lengths = [1061, 2821, 1797, 3822, 22, 1088, 1738]
        expected_dict = dict(zip(ids, gene_lengths))
        assert label_writer._get_mapping_dict_feature_length(ids) == expected_dict

        # Bad
        ids = ["NO_GENE"]
        with pytest.raises(ValueError):
            label_writer._get_mapping_dict_feature_id(ids)

    def test_get_dictionary_mapping_feature_type(self, label_writer):
        # Good
        ids = [
            "ERCC-00002",
            "ENSG00000127603",
            "ENSMUSG00000059552",
            "ENSSASG00005000004",
            "FBtr0472816_df_nrg",
            "ENSDARG00000009657",
            "WBGene00000003",
        ]
        # values derived from csv
        gene_types = [
            "synthetic",
            "protein_coding",
            "protein_coding",
            "protein_coding",
            "ncRNA",
            "protein_coding",
            "protein_coding",
        ]
        expected_dict = dict(zip(ids, gene_types))
        assert label_writer._get_mapping_dict_feature_type(ids) == expected_dict

        # Bad
        ids = ["NO_GENE_BAD"]
        with pytest.raises(ValueError):
            label_writer._get_mapping_dict_feature_type(ids)

    def test_get_dictionary_mapping_feature_biotype(self, label_writer):
        # Good
        ids = [
            "ERCC-00002",
            "ENSG00000127603",
            "ENSMUSG00000059552",
            "ENSSASG00005000004",
            "FBtr0472816_df_nrg",
            "ENSDARG00000009657",
            "WBGene00000003",
        ]
        # Values derived from csv
        biotypes = [
            "spike-in",
            "gene",
            "gene",
            "gene",
            "gene",
            "gene",
            "gene",
        ]
        expected_dict = dict(zip(ids, biotypes))
        assert label_writer._get_mapping_dict_feature_biotype(ids) == expected_dict

    @pytest.mark.parametrize(
        "ids,labels,curie_constraints",
        [
            (["CL:0000066", "CL:0000192"], ["epithelial cell", "smooth muscle cell"], "cell_type_ontology_term_id"),
            (["EFO:0009899", "EFO:0009922"], ["10x 3' v2", "10x 3' v3"], "assay_ontology_term_id"),
            (["MONDO:0100096"], ["COVID-19"], "disease_ontology_term_id"),
        ],
    )
    def test_get_dictionary_mapping_curie__good(self, schema_def, label_writer, ids, labels, curie_constraints):
        # Good
        curie_constraints = schema_def["components"]["obs"]["columns"][curie_constraints]["curie_constraints"]
        expected_dict = dict(zip(ids, labels))
        assert label_writer._get_mapping_dict_curie(ids, curie_constraints) == expected_dict

    def test_get_dictionary_mapping_curie__self_reported_ethnicity_ontology_term_id(self, schema_def, label_writer):
        ids = ["HANCESTRO:0005", "HANCESTRO:0014", "HANCESTRO:0005,HANCESTRO:0014", "unknown"]
        labels = ["European", "Hispanic or Latin American", "European,Hispanic or Latin American", "unknown"]
        curie_constraints = schema_def["components"]["obs"]["columns"]["self_reported_ethnicity_ontology_term_id"][
            "dependencies"
        ][0]["curie_constraints"]
        expected_dict = dict(zip(ids, labels))
        assert label_writer._get_mapping_dict_curie(ids, curie_constraints) == expected_dict

    @pytest.mark.parametrize(
        "ids", [["CL:0000066", "CL:0000192_FOO"], ["CL:0000066", "CL:0000192", "UBERON:0002048"], ["CL:NO_TERM"]]
    )
    def test_get_dictionary_mapping_curie__bad(self, schema_def, label_writer, ids):
        curie_constraints = schema_def["components"]["obs"]["columns"]["cell_type_ontology_term_id"][
            "curie_constraints"
        ]
        with pytest.raises(ValueError):
            label_writer._get_mapping_dict_curie(ids, curie_constraints)

    def test__write__Success(self, label_writer):
        with tempfile.TemporaryDirectory() as temp_dir:
            labels_path = "/".join([temp_dir, "labels.h5ad"])
            assert label_writer.write_labels(labels_path)
        assert not label_writer.errors

    def test__write__Fail(self, label_writer):
        label_writer.adata.write_h5ad = mock.Mock(side_effect=Exception("Test Fail"))
        with tempfile.TemporaryDirectory() as temp_dir:
            labels_path = "/".join([temp_dir, "labels.h5ad"])
            assert not label_writer.write_labels(labels_path)
        assert label_writer.errors


class TestIgnoreLabelFunctions:
    def test_validating_labeled_h5ad_should_fail_if_no_flag_set(self):
        validator = Validator()
        validator.adata = adata_with_labels.copy()
        is_valid = validator.validate_adata()

        assert not is_valid

    def test_validating_labeled_h5ad_should_pass_if_flag_set(self):
        validator = Validator(ignore_labels=True)
        validator.adata = adata_with_labels.copy()
        is_valid = validator.validate_adata()

        assert is_valid


class TestValidate:
    @staticmethod
    def hash_file(file_name: str) -> str:
        with open(file_name, "rb") as f:
            # Read the contents of the file in chunks
            chunk_size = 1024
            hasher = hashlib.sha256()
            while chunk := f.read(chunk_size):
                hasher.update(chunk)
        return hasher.hexdigest()

    def test__validate_with_h5ad_valid_and_labels(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            labels_path = "/".join([temp_dir, "labels.h5ad"])

            success, errors, _ = validate(h5ad_valid, labels_path)

            import anndata as ad

            adata = ad.read_h5ad(labels_path)
            assert adata.X.has_canonical_format
            assert adata.raw.X.has_canonical_format
            assert success
            assert not errors
            assert os.path.exists(labels_path)
            expected_hash = "55fbc095218a01cad33390f534d6690af0ecd6593f27d7cd4d26e91072ea8835"
            original_hash = self.hash_file(h5ad_valid)
            assert original_hash != expected_hash, "Writing labels did not change the dataset from the original."

    def test__validate_with_h5ad_valid_and_without_labels(self):
        success, errors, _ = validate(h5ad_valid)

        assert success
        assert not errors

    def test__validate_with_h5ad_invalid_and_with_labels(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            labels_path = "/".join([temp_dir, "labels.h5ad"])

            success, errors, _ = validate(h5ad_invalid, labels_path)

            assert not success
            assert errors
            assert not os.path.exists(labels_path)

    def test__validate_with_h5ad_invalid_and_without_labels(self):
        success, errors, _ = validate(h5ad_invalid)

        assert not success
        assert errors


class TestCheckSpatial:
    @pytest.mark.parametrize(
        "assay_ontology_term_id, expected_is_visium",
        [
            # Parent term for Visium Spatial Gene Expression. This term and all its descendants are Visium
            ("EFO:0022858", True),
            # Visium Spatial Gene Expression V1
            ("EFO:0022857", True),
            # Visium CytAssist Spatial Gene Expression V2
            ("EFO:0022858", True),
            # Visium CytAssist Spatial Gene Expression, 11mm
            ("EFO:0022860", True),
            # Visium CytAssist Spatial Gene Expression, 6.5mm
            ("EFO:0022859", True),
            # Random other EFO term
            ("EFO:0003740", False),
        ],
    )
    def test__is_visium_descendant(self, assay_ontology_term_id, expected_is_visium):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id

        assert validator._is_visium_including_descendants() == expected_is_visium

    def test__validate_spatial_visium_ok(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator._visium_and_is_single_true_matrix_size = 2
        # Confirm spatial is valid.
        validator.validate_adata()
        assert not validator.errors

    def test__forbid_generic_visium(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator._visium_and_is_single_true_matrix_size = 2

        # set assay to the generic visium term
        validator.adata.obs["assay_ontology_term_id"] = "EFO:0010961"

        # Confirm this triggers FORBIDDIN ERROR and downstream errors due to invalid spatial term.
        validator.validate_adata()
        EXPECTED_FORBIDDEN_ERROR = "ERROR: Invalid spatial assay. obs['assay_ontology_term_id'] must be a descendant of EFO:0010961 but NOT EFO:0010961 itself. "
        assert len(validator.errors) == 5
        assert EXPECTED_FORBIDDEN_ERROR in validator.errors

    @mock.patch("cellxgene_schema.validate.VISIUM_AND_IS_SINGLE_TRUE_MATRIX_SIZE", 2)
    def test__validate_from_file(self):
        """Testing compatibility with SparseDataset types in Anndata"""
        validator: Validator = Validator()
        validator._set_schema_def()
        validator._visium_and_is_single_true_matrix_size = 2
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = os.path.join(temp_dir, "visium.h5ad")
            adata_visium.write_h5ad(file_path)
            # Confirm spatial is valid.
            validator.validate_adata(file_path)
        assert not validator.errors

    def test__validate_spatial_visium_dense_matrix_ok(self):
        """
        Test visium specific requirements on a dense X matrix
        """
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator._visium_and_is_single_true_matrix_size = 2
        _Xdense = validator.adata.X.compute()
        _Xdense[_Xdense == 0] = 1  # ensure the matrix doesn't trigger sparsity error
        validator.adata.X = from_array(_Xdense.toarray())  # daskify
        validator.adata.raw = validator.adata.copy()
        validator.adata.raw.var.drop("feature_is_filtered", axis=1, inplace=True)
        # Confirm spatial is valid.
        validator.validate_adata()
        assert not validator.errors

    def test__validate_spatial_visium_and_is_single_false_ok(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"] = {"is_single": False}
        del validator.adata.obsm["spatial"]
        # Format adata.obs into valid shape for Visium and is_single False.
        validator.adata.obs.pop("array_col")
        validator.adata.obs.pop("array_row")
        validator.adata.obs.pop("in_tissue")
        validator.adata.obs["is_primary_data"] = False
        # Confirm spatial is valid.
        validator.validate_adata()
        assert not validator.errors

    def test__validate_spatial_slide_seqV2_ok(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_slide_seqv2.copy()

        # Confirm spatial is valid.
        validator.validate_adata()
        assert not validator.errors

    @pytest.mark.parametrize("spatial", [None, "invalid", 1, 1.0, True])
    def test__validate_spatial_type_error(self, spatial):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"] = spatial

        # Confirm key type dict is required.
        validator.validate_adata()
        assert (
            validator.errors[0]
            == "ERROR: A dict in uns['spatial'] is required when obs['assay_ontology_term_id'] is either a descendant of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2)."
        )

    def test__validate_spatial_is_single_false_ok(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_spatial_is_single_false.copy()

        # Confirm spatial is valid.
        validator.validate_adata()
        assert not validator.errors

    def test__validate_spatial_forbidden_if_not_visium_or_slide_seqv2(self):
        validator: Validator = Validator()
        validator._set_schema_def()

        # Create Visium adata (with spatial) with 10x 3' v2 obs.
        validator.adata = adata_visium.copy()
        validator.adata.obs = good_obs.copy()

        # Confirm spatial is not allowed for 10x 3' v2.
        validator._check_spatial_uns()
        assert validator.errors == [
            "uns['spatial'] is only allowed when obs['assay_ontology_term_id'] is either "
            "a descendant of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2)"
        ]

    @pytest.mark.parametrize(
        "assay_ontology_term_id, is_descendant",
        [("EFO:0022859", True), ("EFO:0022858", True), ("EFO:0030029", False), ("EFO:0002697", False)],
    )
    def test__validate_spatial_required_if_visium(self, assay_ontology_term_id, is_descendant):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id

        if is_descendant:
            # check pass if 'spatial' included
            validator.adata.uns = good_uns_with_visium_spatial.copy()
            validator._check_spatial_uns()
            assert len(validator.errors) == 0
            validator.reset()

            # check fail if 'spatial' not included
            validator.adata.uns = good_uns.copy()
            validator._check_spatial_uns()
            assert validator.errors == [
                "A dict in uns['spatial'] is required when obs['assay_ontology_term_id'] is "
                "either a descendant of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2)."
            ]
            validator.reset()
        else:
            # check fail if 'spatial' included
            validator.adata.uns = good_uns_with_visium_spatial.copy()
            validator._check_spatial_uns()
            assert len(validator.errors) == 1
            validator.reset()

    def test__validate_spatial_required_if_slide_seqV2(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_slide_seqv2.copy()
        validator.adata.uns = good_uns.copy()

        # Confirm spatial is required for Slide-seqV2.
        validator._check_spatial_uns()
        assert validator.errors == [
            "A dict in uns['spatial'] is required when obs['assay_ontology_term_id'] is either a descendant of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2)."
        ]

    def test__validate_spatial_allowed_keys_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"]["invalid_key"] = True

        # Confirm additional key is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'] must contain only two top-level keys: 'is_single' and a library_id. "
            "More than two top-level keys detected:" in validator.errors[0]
        )

    @pytest.mark.parametrize(
        "assay_ontology_term_id, is_descendant",
        [("EFO:0022859", True), ("EFO:0022858", True), ("EFO:0030029", False), ("EFO:0002697", False)],
    )
    def test__validate_is_single_required_visium_error(self, assay_ontology_term_id, is_descendant):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.uns["spatial"].pop("is_single")
        validator._check_spatial_uns()

        if is_descendant:
            # if spatial, MUST specify `is_single`
            assert "uns['spatial'] must contain the key 'is_single'." in validator.errors[0]
        else:
            # if not spatial, MUST NOT speciffy `is_single`
            assert validator.errors == [
                "uns['spatial'] is only allowed when obs['assay_ontology_term_id'] is either a descendant of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2)"
            ]

    def test__validate_is_single_required_slide_seqV2_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()

        # Remove is_single key to trigger error, but add dummy library_id to pass earlier truthy check
        # on spatial existing.
        validator.adata = adata_slide_seqv2.copy()
        validator.adata.uns["spatial"] = {
            "library_id": "test_library_id",
        }

        # Confirm is_single is identified as required.
        validator._check_spatial_uns()
        assert validator.errors
        assert "uns['spatial'] must contain the key 'is_single'." in validator.errors[0]

    def test__validate_is_single_boolean_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"]["is_single"] = "invalid"

        # Confirm is_single is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert "uns['spatial']['is_single'] must be of boolean type" in validator.errors[0]

    def test__validate_library_id_forbidden_if_slide_seqV2(self):
        validator: Validator = Validator()
        validator._set_schema_def()

        # Add library_id to Slide-seqV2 uns to trigger error.
        validator.adata = adata_slide_seqv2.copy()
        validator.adata.uns = good_uns_with_visium_spatial

        # Confirm library_id is not allowed for Slide-seqV2.
        validator._check_spatial_uns()
        assert len(validator.errors) == 1
        assert f"uns['spatial'][library_id] {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_FORBIDDEN}." in validator.errors[0]

    def test__validate_library_id_forbidden_if_visium_or_is_single_false(self):
        validator: Validator = Validator()
        validator._set_schema_def()

        # Set is_single to False to trigger error.
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"]["is_single"] = np.bool_(False)

        # Confirm library_id is not allowed for Visium if is_single is False.
        validator._check_spatial_uns()
        assert len(validator.errors) == 1
        assert f"uns['spatial'][library_id] {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_FORBIDDEN}." in validator.errors[0]

    @pytest.mark.parametrize(
        "assay_ontology_term_id, is_descendant",
        [("EFO:0022859", True), ("EFO:0022858", True), ("EFO:0030029", False), ("EFO:0002697", False)],
    )
    def test__validate_library_id_required_if_visium(self, assay_ontology_term_id, is_descendant):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()

        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        if is_descendant:
            # if spatial, `library_id` must exist
            validator._check_spatial_uns()
            assert len(validator.errors) == 0
            validator.reset()

            # if spatial, but missing from `uns`
            validator.adata.uns["spatial"].pop(visium_library_id)
            validator._check_spatial_uns()
            assert validator.errors == [
                f"uns['spatial'] must contain at least one key representing the library_id when {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE}."
            ]
        else:
            # if not spatial, MUST NOT define `library_id`
            validator.adata.uns["spatial"][visium_library_id] = {"images": []}
            validator._check_spatial_uns()
            # Report the most general top level error
            assert validator.errors == [
                "uns['spatial'] is only allowed when obs['assay_ontology_term_id'] is either a descendant of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2)"
            ]

    @pytest.mark.parametrize("library_id", [None, "invalid", 1, 1.0, True])
    def test__validate_library_id_type_error(self, library_id):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id] = library_id

        # Confirm library_id is identified as invalid.
        validator.validate_adata()
        assert validator.errors
        assert "uns['spatial'][library_id] must be a dictionary." in validator.errors[0]

    def test__validate_library_id_allowed_keys_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["invalid_key"] = True

        # Confirm additional key is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id] can only contain the keys 'images' and 'scalefactors'." in validator.errors[0]
        )

    def test__validate_images_required_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id].pop("images")

        # Confirm images is required.
        validator._check_spatial_uns()
        assert validator.errors
        assert "uns['spatial'][library_id] must contain the key 'images'." in validator.errors[0]

    @pytest.mark.parametrize(
        "assay_ontology_term_id, is_descendant",
        [("EFO:0010961", True), ("EFO:0022858", True), ("EFO:0030029", False), ("EFO:0002697", False)],
    )
    def test__validate_images_allowed_keys_error(self, assay_ontology_term_id, is_descendant):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["images"]["invalid_key"] = True

        # Confirm additional key is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id]['images'] can only contain the keys 'fullres' and 'hires'."
            in validator.errors[0]
        )

    def test__validate_images_hires_required_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["images"].pop("hires")

        # Confirm hires is required.
        validator._check_spatial_uns()
        assert validator.errors
        assert "uns['spatial'][library_id]['images'] must contain the key 'hires'." in validator.errors[0]

    def test__validate_images_fullres_optional(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["images"].pop("fullres")

        # Confirm fullres is optional.
        validator._check_spatial_uns()
        assert not validator.errors
        assert validator.warnings
        assert (
            "No uns['spatial'][library_id]['images']['fullres'] was found. "
            "It is STRONGLY RECOMMENDED that uns['spatial'][library_id]['images']['fullres'] is provided."
            in validator.warnings[0]
        )

    @pytest.mark.parametrize(
        "image_name, image_shape",
        [
            ("hires", (1, 2000, 4)),
            ("fullres", (1, 1, 4)),
        ],
    )
    def test__validate_images_image_last_dimension_4_ok(self, image_name, image_shape):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["images"][image_name] = np.zeros(image_shape, dtype=np.uint8)

        # Confirm image is valid.
        validator._check_spatial_uns()
        assert not validator.errors

    @pytest.mark.parametrize(
        "image_name, image_shape",
        [
            ("hires", (1, 2000, 3)),
            ("fullres", (1, 1, 3)),
        ],
    )
    def test__validate_images_image_ndarray_type_error(self, image_name, image_shape):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        # Defaults to float64.
        validator.adata.uns["spatial"][visium_library_id]["images"][image_name] = np.zeros(image_shape)

        # Confirm image is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            f"uns['spatial'][library_id]['images']['{image_name}'] must be of type numpy.uint8" in validator.errors[0]
        )

    @pytest.mark.parametrize(
        "image_name",
        [
            "hires",
            "fullres",
        ],
    )
    def test__validate_images_image_is_ndarray_error(self, image_name):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["images"][image_name] = "invalid"

        # Confirm image is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            f"uns['spatial'][library_id]['images']['{image_name}'] must be of numpy.ndarray type" in validator.errors[0]
        )

    @pytest.mark.parametrize(
        "image_name",
        [
            "hires",
            "fullres",
        ],
    )
    def test__validate_images_image_is_shape_error(self, image_name):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["images"][image_name] = np.zeros((1, 1), dtype=np.uint8)

        # Confirm image is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            f"uns['spatial'][library_id]['images']['{image_name}'] must have a length of 3 and either 3 (RGB color model "
            "for example) or 4 (RGBA color model for example) for its last dimension" in validator.errors[0]
        )

    @pytest.mark.parametrize(
        "assay_ontology_term_id, hi_res_size, image_max",
        [
            ("EFO:0022858", 2001, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE),
            ("EFO:0022860", 4001, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM),
        ],
    )
    def test__validate_images_hires_max_dimension_greater_than_error(
        self, assay_ontology_term_id, hi_res_size, image_max
    ):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.uns["spatial"][visium_library_id]["images"]["hires"] = np.zeros(
            (1, hi_res_size, 3), dtype=np.uint8
        )

        # Confirm hires is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors == [
            f"The largest dimension of uns['spatial'][library_id]['images']['hires'] must be {image_max} pixels, it has a largest dimension of {hi_res_size} pixels."
        ]

    @pytest.mark.parametrize(
        "assay_ontology_term_id, hi_res_size, size_requirement",
        [
            ("EFO:0022858", SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE),
            ("EFO:0022858", SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE),
            ("EFO:0022860", SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM),
            (
                "EFO:0022860",
                SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM,
                SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM,
            ),
        ],
    )
    def test__validate_images_hires_max_dimension(self, assay_ontology_term_id, hi_res_size, size_requirement):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.uns["spatial"][visium_library_id]["images"]["hires"] = np.zeros(
            (1, hi_res_size, 3), dtype=np.uint8
        )

        # Confirm hires is identified as invalid.
        validator.reset()
        validator._check_spatial_uns()
        if hi_res_size == size_requirement:
            assert validator.errors == []
        else:
            assert validator.errors == [
                f"The largest dimension of uns['spatial'][library_id]['images']['hires'] must be {size_requirement} pixels, it has a largest dimension of {hi_res_size} pixels."
            ]

    @pytest.mark.parametrize(
        "assay_ontology_term_id, hi_res_size, image_max",
        [
            ("EFO:0022858", 1999, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE),
            ("EFO:0022860", 3999, SPATIAL_HIRES_IMAGE_MAX_DIMENSION_SIZE_VISIUM_11MM),
        ],
    )
    def test__validate_images_hires_max_dimension_less_than_error(self, assay_ontology_term_id, hi_res_size, image_max):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.uns["spatial"][visium_library_id]["images"]["hires"] = np.zeros(
            (1, hi_res_size, 3), dtype=np.uint8
        )

        # Confirm hires is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors == [
            f"The largest dimension of uns['spatial'][library_id]['images']['hires'] must be {image_max} pixels, it has a largest dimension of {hi_res_size} pixels."
        ]

    def test__validate_scalefactors_required_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id].pop("scalefactors")

        # Confirm scalefactors is required.
        validator._check_spatial_uns()
        assert validator.errors
        assert "uns['spatial'][library_id] must contain the key 'scalefactors'." in validator.errors[0]

    def test__validate_scalefactors_allowed_keys_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["scalefactors"]["invalid_key"] = True

        # Confirm additional key is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id]['scalefactors'] can only contain the keys "
            "'spot_diameter_fullres' and 'tissue_hires_scalef'." in validator.errors[0]
        )

    def test__validate_scalefactors_spot_diameter_fullres_required_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["scalefactors"].pop("spot_diameter_fullres")

        # Confirm spot_diameter_fullres is required.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id]['scalefactors'] must contain the key 'spot_diameter_fullres'."
            in validator.errors[0]
        )

    def test__validate_scalefactors_tissue_hires_scalef_required_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["scalefactors"].pop("tissue_hires_scalef")

        # Confirm tissue_hires_scalef is required.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id]['scalefactors'] must contain the key 'tissue_hires_scalef'."
            in validator.errors[0]
        )

    def test__validate_scalefactors_spot_diameter_fullres_is_float_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["scalefactors"]["spot_diameter_fullres"] = "invalid"

        # Confirm spot_diameter_fullres is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id]['scalefactors']['spot_diameter_fullres'] must be of type float"
            in validator.errors[0]
        )

    def test__validate_scalefactors_tissue_hires_scalef_is_float_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id]["scalefactors"]["tissue_hires_scalef"] = "invalid"

        # Confirm tissue_hires_scalef is identified as invalid.
        validator._check_spatial_uns()
        assert validator.errors
        assert (
            "uns['spatial'][library_id]['scalefactors']['tissue_hires_scalef'] must be of type float"
            in validator.errors[0]
        )

    @pytest.mark.parametrize("key", ["scalefactors", "images"])
    def test__validate_library_id_key_value_type_error(self, key):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.uns["spatial"][visium_library_id][key] = "invalid"

        # Confirm key type dict is required.
        validator._check_spatial_uns()
        assert validator.errors
        assert f"uns['spatial'][library_id]['{key}'] must be a dictionary." in validator.errors[0]

    def test__validate_assay_type_ontology_term_id_not_unique_error(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs.assay_ontology_term_id = ["EFO:0022858", "EFO:0030062"]

        # Confirm assay ontology term id is identified as invalid.
        validator._validate_spatial_assay_ontology_term_id()
        assert validator.errors
        assert (
            "When obs['assay_ontology_term_id'] is either a descendant"
            " of 'EFO:0010961' (Visium Spatial Gene Expression) or 'EFO:0030062' (Slide-seqV2), all observations must contain the same value."
        ) in validator.errors[0]

    def test__validate_assay_type_ontology_term_id_not_unique_ok(self, valid_adata):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = valid_adata  # "EFO:0009899" and "EFO:0009918"

        # Confirm assay ontology term id is considered valid.
        validator._validate_spatial_assay_ontology_term_id()
        assert not validator.errors

    @pytest.mark.parametrize(
        "assay_ontology_term_id, is_single",
        [
            (["EFO:0022858", "EFO:0030062"], True),
            (["EFO:0022858", "EFO:0030062"], False),
            ("EFO:0022858", False),
            ("EFO:0030062", True),
            ("EFO:0030062", False),
            ("EFO:0030062", False),
            ("EFO:0009899", True),  # Non-spatial
            ("EFO:0009899", False),  #  Non-spatial
        ],
    )
    def test__validate_tissue_position_forbidden(self, assay_ontology_term_id, is_single):
        validator: Validator = Validator()
        validator._set_schema_def()

        # Create Visium adata with tissue positions, update assay_ontology_term_id and is_single to trigger error.
        validator.adata = adata_visium.copy()
        validator.adata.obs.assay_ontology_term_id = assay_ontology_term_id
        validator.adata.uns["spatial"]["is_single"] = is_single
        validator.adata.obs["is_primary_data"] = False

        # Confirm tissue positions are not allowed.
        validator._validate_spatial_tissue_positions()
        assert len(validator.errors) == 3
        tissue_position_names = ["array_col", "array_row", "in_tissue"]
        for i, tissue_position_name in enumerate(tissue_position_names):
            assert (
                f"obs['{tissue_position_name}'] {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_FORBIDDEN}."
                in validator.errors[i]
            )

    @pytest.mark.parametrize("tissue_position_name", ["array_col", "array_row", "in_tissue"])
    def test__validate_tissue_position_required(self, tissue_position_name):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs.pop(tissue_position_name)

        # check visium
        validator.adata.obs["assay_ontology_term_id"] = "EFO:0022858"
        validator._check_spatial_obs()
        assert validator.errors
        assert (
            f"obs['{tissue_position_name}'] {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_REQUIRED}." in validator.errors[0]
        )
        validator.reset()

        # check visium descendant
        validator.adata.obs["assay_ontology_term_id"] = "EFO:0022860"
        validator._check_spatial_obs()
        assert validator.errors
        assert (
            f"obs['{tissue_position_name}'] {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_REQUIRED}." in validator.errors[0]
        )
        validator.reset()

    @pytest.mark.parametrize("assay_ontology_term_id", ["EFO:0022858", "EFO:0030062", "EFO:0022860"])
    def test__validate_tissue_position_not_required(self, assay_ontology_term_id):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_slide_seqv2.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.uns["spatial"]["is_single"] = False  # setting to false removes the requirement
        validator.adata.obs["is_primary_data"] = False
        validator._check_spatial_obs()
        assert not validator.errors

    @pytest.mark.parametrize("tissue_position_name", ["array_col", "array_row", "in_tissue"])
    def test__validate_tissue_position_int_error(self, tissue_position_name):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs[tissue_position_name] = 1.0

        # Confirm tissue_position is identified as invalid.
        validator._check_spatial_obs()
        assert validator.errors
        assert f"obs['{tissue_position_name}'] must be of int type" in validator.errors[0]

    @pytest.mark.parametrize("tissue_position_name", ["array_col", "array_row", "in_tissue"])
    def test__validate_tissue_position_nan_error(self, tissue_position_name):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs[tissue_position_name] = np.nan

        # Confirm tissue_position is identified as invalid.
        validator._check_spatial_obs()
        assert validator.errors[0] == f"obs['{tissue_position_name}'] {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_NOTNULL}."

    @pytest.mark.parametrize("assay_ontology_term_id", ["EFO:0022857", "EFO:0022860", "EFO:0022859"])
    @pytest.mark.parametrize("tissue_position_name, min", [("array_col", 0), ("array_row", 0), ("in_tissue", 0)])
    def test__validate_tissue_position_int_min_error(self, assay_ontology_term_id, tissue_position_name, min):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.obs[tissue_position_name] = min - 1

        # Confirm tissue_position is identified as invalid.
        validator._check_spatial_obs()
        assert (
            re.match(f"^obs\['{tissue_position_name}'\] must be (between )?{min} (and|or) [0-9]+", validator.errors[0])
            is not None
        )

    @pytest.mark.parametrize(
        "assay_ontology_term_id, tissue_position_name, tissue_position_max",
        [
            ("EFO:0022857", "array_col", 127),
            ("EFO:0022857", "array_row", 77),
            ("EFO:0022860", "array_col", 223),
            ("EFO:0022860", "array_row", 127),
            ("EFO:0022859", "array_col", 127),
            ("EFO:0022859", "array_row", 77),
            ("EFO:0022859", "in_tissue", 1),
        ],
    )
    def test__validate_tissue_position_int_max_error(
        self, assay_ontology_term_id, tissue_position_name, tissue_position_max
    ):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        validator.adata.obs[tissue_position_name] = tissue_position_max + 1

        # Confirm tissue_position is identified as invalid.
        validator._check_spatial_obs()
        assert (
            re.match(
                f"^obs\['{tissue_position_name}'\] must be (between )?[0-9]+ (and|or) {tissue_position_max}",
                validator.errors[0],
            )
            is not None
        )

    @pytest.mark.parametrize(
        "cell_type_ontology_term_id, in_tissue, assay_ontology_term_id",
        [
            # MUST be unknown when in_tissue = 0 and assay_ontology_term_id = Visium Spatial Gene Expression v2
            ("unknown", 0, "EFO:0022858"),
            # MUST be unknown when in_tissue = 0 and assay_ontology_term_id = Visium CytAssist Spatial Gene Expression, 11mm
            ("unknown", 0, "EFO:0022860"),
            # MUST be unknown when in_tissue = 0 and assay_ontology_term_id = Visium Spatial Gene Expression V1
            # valid CL term is ok when in_tissue = 1 and assay_ontology_term_id = Visium CytAssist Spatial Gene Expression, 11mm
            (["unknown", "CL:0000066"], [0, 1], ["EFO:0022857", "EFO:0022860"]),
            # normal CL term for in_tissue = 1 and assay_ontology_term_id = 10x 3' v2
            ("CL:0000066", 1, "EFO:0009899"),
        ],
    )
    def test__validate_cell_type_ontology_term_id_ok(
        self, cell_type_ontology_term_id, in_tissue, assay_ontology_term_id
    ):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs.cell_type_ontology_term_id = cell_type_ontology_term_id
        validator.adata.obs.in_tissue = in_tissue
        validator.adata.obs.assay_ontology_term_id = assay_ontology_term_id

        # Confirm cell type is valid.
        validator._validate_spatial_cell_type_ontology_term_id()
        assert not validator.errors

    @pytest.mark.parametrize(
        "cell_type_ontology_term_id, in_tissue, assay_ontology_term_id",
        [
            # MUST be unknown when in_tissue = 0 and assay_ontology_term_id = Visium Spatial Gene Expression
            ("CL:0000066", 0, "EFO:0022858"),
            (["CL:0000066", "unknown"], [0, 1], ["EFO:0022858", "EFO:0022858"]),
            # MUST be unknown when in_tissue = 0 and assay_ontology_term_id = Visium CytAssist Spatial Gene Expression, 11mm
            ("CL:0000066", 0, "EFO:0022860"),
            # MUST be unknown when in_tissue = 0 and assay_ontology_term_id = Visium Spatial Gene Expression V1
            ("CL:0000066", 0, "EFO:0022857"),
        ],
    )
    def test__validate_cell_type_ontology_term_id_error(
        self, cell_type_ontology_term_id, in_tissue, assay_ontology_term_id
    ):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator.adata.obs.cell_type_ontology_term_id = cell_type_ontology_term_id
        validator.adata.obs.in_tissue = in_tissue
        validator.adata.obs.assay_ontology_term_id = assay_ontology_term_id

        # Confirm errors.
        validator._validate_spatial_cell_type_ontology_term_id()
        assert validator.errors
        assert (
            f"obs['cell_type_ontology_term_id'] must be 'unknown' when {ERROR_SUFFIX_VISIUM_AND_IS_SINGLE_TRUE_IN_TISSUE_0}."
            in validator.errors[0]
        )

    def test__validate_embeddings_non_nans(self):
        validator: Validator = Validator()
        validator._set_schema_def()
        validator.adata = adata_visium.copy()
        validator._visium_and_is_single_true_matrix_size = 2

        # invalidate spatial embeddings with NaN value
        validator.adata.obsm["spatial"][0, 1] = np.nan
        # Confirm spatial is valid.
        validator.validate_adata()
        assert validator.errors == ["ERROR: adata.obsm['spatial'] contains at least one NaN value."]


class TestValidatorValidateDataFrame:
    @pytest.mark.parametrize("_type", [np.int64, np.int32, int, np.float64, np.float32, float, str])
    def test_succeed_categorical_types(self, tmp_path, _type, valid_adata):
        # Arrange
        categories = [*map(_type, range(adata_valid.n_obs))]
        self._add_catagorical_obs(valid_adata, categories)
        validator = self._create_validator(valid_adata)

        # Act
        validator._validate_dataframe("obs")

        # Assert
        assert not validator.errors
        valid_adata.write_h5ad(f"{tmp_path}/test.h5ad")  # Succeed write

    def test_fail_categorical_mixed_types(self, tmp_path, valid_adata):
        # Arrange
        categories = ["hello", 123]
        self._add_catagorical_obs(valid_adata, categories)
        validator = self._create_validator(valid_adata)

        # Act
        validator._validate_dataframe("obs")

        # Assert
        assert "in dataframe 'obs' contains 2 categorical types. Only one type is allowed." in validator.errors[0]
        self._fail_write_h5ad(tmp_path, valid_adata)

    def _add_catagorical_obs(self, adata, categories):
        t = pd.CategoricalDtype(categories=categories)
        adata.obs["test_cat"] = pd.Series(data=categories, index=["X", "Y"], dtype=t)

    def _create_validator(self, adata):
        validator = Validator()
        validator._set_schema_def()
        validator.adata = adata
        return validator

    def _fail_write_h5ad(self, tmp_path, adata):
        with pytest.raises(TypeError):
            # If this tests starts to fail here it means the anndata version has be upgraded and this check is no
            # longer needed
            adata.write_h5ad(f"{tmp_path}/test.h5ad")

    def test_fail_mixed_column_types(self, tmp_path, valid_adata):
        # Arrange
        valid_adata.obs["mixed"] = pd.Series(data=["1234", 0], index=["X", "Y"])
        validator = self._create_validator(valid_adata)

        # Act
        validator._validate_dataframe("obs")

        # Assert
        assert "in dataframe 'obs' cannot contain mixed types." in validator.errors[0]
        self._fail_write_h5ad(tmp_path, valid_adata)


class TestIsRaw:
    @staticmethod
    def create_validator(data: Union[ndarray, spmatrix], matrix_format: str) -> Validator:
        """
        Create a sample AnnData instance with the given data and format.

        :param data: The data matrix.
        :param matrix_format: The format of the data matrix (e.g., "dense", "csr", "csc").

        :return anndata.AnnData: An AnnData instance with the specified data and format.
        """
        validator = Validator()

        adata = anndata.AnnData(X=data)
        adata.obsm["X_" + matrix_format] = data

        validator.adata = adata
        return validator

    @pytest.mark.parametrize(
        "data, matrix_format, expected_result",
        [
            # Test case with integer values in a dense matrix
            (from_array(np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)), "dense", True),
            # Test case with float values in a dense matrix
            (from_array(np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])), "dense", False),
            # Test case with integer values in a sparse matrix (CSR format)
            (from_array(sparse.csr_matrix([[1, 0, 3], [0, 5, 0]], dtype=np.float32)), "csr", True),
            # Test case with float values in a sparse matrix (CSC format)
            (from_array(sparse.csc_matrix([[1.1, 0, 3.3], [0, 5.5, 0]])), "csc", False),
            # Test case with mixed integer and float values in a dense matrix
            (from_array(np.array([[1, 2.2, 3], [4.4, 5, 6.6]])), "dense", False),
        ],
    )
    def test_has_valid_raw(self, data, matrix_format, expected_result):
        validator = self.create_validator(data, matrix_format)
        assert validator._has_valid_raw() == expected_result

    @mock.patch("cellxgene_schema.validate.get_matrix_format", return_value="unknown")
    def test_has_valid_raw_with_unknown_format(self, mock_get_matrix_format):
        # a matrix with unknown format should be invalid
        data = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float32)
        validator = self.create_validator(data, "unknown")
        assert validator._has_valid_raw() is False
        assert validator.errors == [f"Unknown encoding for matrix X. {ERROR_SUFFIX_SPARSE_FORMAT}"]


class TestValidateAnndataRawCounts:
    """
    Paired Assays (measures both accessibility and gene expression) require raw counts to be present and validated.
    Unpaired Assays (measures only accessibility) do not require raw counts to be present.

    This is the purview of the AnnData Validator, not the AtacValidator, which is primarily focused on the fragment file.
    """

    def test_paired_requires_raw_validation(self, atac_anndata, tmpdir):
        # 10x multiome (EFO:0030059) is paired (both ATAC and RNA single cell sequencing)
        atac_anndata.obs["assay_ontology_term_id"] = ["EFO:0030059"] * 3

        # use a valid count matrix (as dask array)
        X = atac_anndata.X

        # validate with AnnData Validator
        validator = Validator(ignore_labels=True)
        validator._set_schema_def()

        # do validation with a valid count matrix
        atac_anndata.X = from_array(X.astype("float32"))
        validator.adata = atac_anndata
        validator.reset()
        validator._validate_raw()
        assert validator.errors == []

        # do validation with an invalid count matrix
        atac_anndata.X = from_array(zeros(X.shape).astype("float32"))
        validator.adata = atac_anndata
        validator.reset()
        validator._validate_raw()
        assert len(validator.errors) > 0

    def test_unpaired_skips_raw_validation(self, atac_anndata, tmpdir):
        # scATAC-seq (EFO:0010891) is unpaired paired
        atac_anndata.obs["assay_ontology_term_id"] = ["EFO:0010891"] * 3

        # remove matrix - it shouldn't be required
        del atac_anndata.X

        # check that validation passes even without matrix
        validator = Validator(ignore_labels=True)
        validator._set_schema_def()
        validator.adata = atac_anndata
        validator.reset()
        validator._validate_raw()
        assert validator.errors == []



import pytest
from cellxgene_schema import gencode
from fixtures.examples_ontology_test import (
    invalid_genes,
    invalid_species,
    valid_genes,
    valid_genes_same_name_and_species,
    valid_genes_same_name_diff_species,
)

# Tests for internal functions of the OntologyChecker and GeneChecker classes


class TestGeneChecker:
    @pytest.mark.parametrize("valid_species", gencode.SupportedOrganisms)
    def test_species_valid(self, valid_species):
        assert isinstance(gencode.GeneChecker(valid_species), gencode.GeneChecker)

    @pytest.mark.parametrize("invalid_species", invalid_species)
    def test_species_invalid(self, invalid_species):
        with pytest.raises(ValueError):
            gencode.GeneChecker(invalid_species)

    @pytest.mark.parametrize("species,valid_genes", valid_genes.items())
    def test_valid_genes(self, species, valid_genes):
        geneChecker = gencode.GeneChecker(species)
        for gene_id in valid_genes:
            gene_label = valid_genes[gene_id][0]
            gene_length = valid_genes[gene_id][1]

            assert geneChecker.is_valid_id(gene_id)
            assert geneChecker.get_symbol(gene_id) == gene_label
            assert geneChecker.get_length(gene_id) == gene_length
            assert gencode.get_organism_from_feature_id(gene_id) == species

    @pytest.mark.parametrize("species,invalid_genes", invalid_genes.items())
    def test_invalid_genes(self, species, invalid_genes):
        geneChecker = gencode.GeneChecker(species)
        for gene_id in invalid_genes:
            assert not geneChecker.is_valid_id(gene_id)
            with pytest.raises(ValueError):
                geneChecker.get_symbol(gene_id)
            with pytest.raises(ValueError):
                geneChecker.get_length(gene_id)

    @pytest.mark.parametrize("species,valid_genes_same_name_diff_species", valid_genes_same_name_diff_species.items())
    def test_valid_genes_same_name_diff_species(self, species, valid_genes_same_name_diff_species):
        geneChecker = gencode.GeneChecker(species)
        for gene_id in valid_genes_same_name_diff_species:
            gene_label = valid_genes_same_name_diff_species[gene_id][0]
            gene_length = valid_genes_same_name_diff_species[gene_id][1]

            assert geneChecker.is_valid_id(gene_id)
            assert geneChecker.get_symbol(gene_id) == gene_label
            assert geneChecker.get_length(gene_id) == gene_length

    @pytest.mark.parametrize("species,valid_genes_same_name_and_species", valid_genes_same_name_and_species.items())
    def test_valid_genes_same_name_and_species(self, species, valid_genes_same_name_and_species):
        geneChecker = gencode.GeneChecker(species)
        for gene_id in valid_genes_same_name_and_species:
            gene_label = valid_genes_same_name_and_species[gene_id][0]
            gene_length = valid_genes_same_name_and_species[gene_id][1]

            assert geneChecker.is_valid_id(gene_id)
            assert geneChecker.get_symbol(gene_id) == gene_label
            assert geneChecker.get_length(gene_id) == gene_length



import logging
import os
import sys
from base64 import b85encode
from functools import lru_cache
from typing import Dict, List, Union

import anndata as ad
import h5py
import numpy as np
from anndata.compat import DaskArray
from anndata.experimental import read_dispatched, read_elem_as_dask
from cellxgene_ontology_guide.ontology_parser import OntologyParser
from dask.array import map_blocks
from scipy import sparse
from xxhash import xxh3_64_intdigest

logger = logging.getLogger(__name__)

SPARSE_MATRIX_TYPES = {"csr", "csc", "coo"}
SUPPORTED_SPARSE_MATRIX_TYPES = {"csr"}

KB = 1024
MB = 1024 * KB
GB = 1024 * MB


def replace_ontology_term(dataframe, ontology_name, update_map):
    column_name = f"{ontology_name}_ontology_term_id"
    if dataframe[column_name].dtype != "category":
        dataframe[column_name] = dataframe[column_name].astype("category")
    for old_term, new_term in update_map.items():
        if old_term in dataframe[column_name].cat.categories:
            # add new one if not already in category, else continue
            if new_term not in dataframe[column_name].cat.categories:
                dataframe[column_name] = dataframe[column_name].cat.add_categories(new_term)
            # replace in dataset
            dataframe.loc[dataframe[column_name] == old_term, column_name] = new_term
            # remove deprecated_term from category
            dataframe[column_name] = dataframe[column_name].cat.remove_categories(old_term)


def map_ontology_term(dataframe, ontology_name, map_from_column, update_map):
    column_name = f"{ontology_name}_ontology_term_id"
    if dataframe[column_name].dtype != "category":
        dataframe[column_name] = dataframe[column_name].astype("category")
    for map_value, new_term in update_map.items():
        if new_term not in dataframe[column_name].cat.categories:
            dataframe[column_name] = dataframe[column_name].cat.add_categories(new_term)
        dataframe.loc[dataframe[map_from_column] == map_value, column_name] = new_term
    dataframe[column_name] = dataframe[column_name].cat.remove_unused_categories()


def remove_deprecated_features(*, adata: ad.AnnData, deprecated: List[str]) -> ad.AnnData:
    # Filter out genes that don't appear in the approved annotation
    var_to_keep = adata.var.index[~adata.var.index.isin(deprecated)].tolist()
    adata = adata[:, var_to_keep].copy()

    # Repeat much of the same steps for the raw.var, if it exists
    if adata.raw:
        raw_adata = ad.AnnData(adata.raw.X, var=adata.raw.var, obs=adata.obs)
        var_to_keep = raw_adata.var.index[~raw_adata.var.index.isin(deprecated)].tolist()
        raw_adata = raw_adata[:, var_to_keep].copy()
        adata.raw = raw_adata
    return adata


def remap_deprecated_features(*, adata: ad.AnnData, remapped_features: Dict[str, str]) -> ad.AnnData:
    # Use remapped_terms to map to the new term ids
    adata.var.index = [remapped_features.get(val, val) for val in adata.var.index]

    # Repeat much of the same steps for the raw.var, if it exists
    if adata.raw:
        raw_adata = ad.AnnData(adata.raw.X, var=adata.raw.var, obs=adata.obs)
        raw_adata.var.index = [remapped_features.get(val, val) for val in raw_adata.var.index]
        adata.raw = raw_adata
    return adata


def get_matrix_format(matrix: DaskArray) -> str:
    """
    Given a matrix, returns the format as one of: csc, csr, coo, dense
    or unknown.

    This mimics the scipy.sparse `format` property, but extends it to
    support ndarray and other classes AnnData may proxy the matrix with.
    """

    # Note: the AnnData proxy classes DO support the `format_str` property, but
    # doing a slice seemed safer, if less performant.  Using `format_str`, which
    # currently works, uses private API:
    #
    # >>> return getattr(matrix, "format_str", "dense)
    #
    matrix_format = "unknown"
    matrix_slice = matrix[0:1, 0:1].compute()
    if isinstance(matrix_slice, sparse.spmatrix):
        matrix_format = matrix_slice.format
    elif isinstance(matrix_slice, np.ndarray):
        matrix_format = "dense"
    assert matrix_format in SPARSE_MATRIX_TYPES.union({"unknown", "dense"})
    return matrix_format


def getattr_anndata(adata: ad.AnnData, attr: str = None):
    """
    same as getattr but handles the special case of "raw.var" for an anndata.AndData object

    :param anndata.AnnData adata: the anndata.AnnData object from which to extract an attribute
    :param str attr: name of the attribute to extract

    :return the attribute or none if it does not exist
    """

    if attr == "raw.var":
        if adata.raw:
            return adata.raw.var
        else:
            return None
    else:
        return getattr(adata, attr)


def read_backed(f: h5py.File, chunk_size: int) -> ad.AnnData:
    """
    Read an AnnData object from a h5py.File object, reading in matrices (dense or sparse) as dask arrays. Does not
    read full matrices into memory.

    :param f: h5py.File object
    :param chunk_size: size of chunks to read matrices in
    :return: ad.AnnData object
    """

    def callback(func, elem_name: str, elem, iospec):
        if "/layers" in elem_name or "/uns" in elem_name or elem_name == "/X" or elem_name == "/raw/X":
            if iospec.encoding_type == "csr_matrix":
                n_vars = elem.attrs.get("shape")[1]
                return read_elem_as_dask(elem, chunks=(chunk_size, n_vars))
            elif iospec.encoding_type == "csc_matrix":
                n_obs = elem.attrs.get("shape")[0]
                return read_elem_as_dask(elem, chunks=(n_obs, chunk_size))
            elif iospec.encoding_type == "array" and len(elem.shape) == 2:
                n_vars = elem.shape[1]
                return read_elem_as_dask(elem, chunks=(chunk_size, n_vars))
            else:
                return func(elem)
        else:
            return func(elem)

    adata = read_dispatched(f, callback=callback)

    return adata


def read_h5ad(h5ad_path: Union[str, bytes, os.PathLike], chunk_size: int = 5000) -> ad.AnnData:
    """
    Reads h5ad into adata
    :params Union[str, bytes, os.PathLike] h5ad_path: path to h5ad to read

    :rtype None
    """
    try:
        f = h5py.File(h5ad_path)
        adata = read_backed(f, chunk_size)

    except (OSError, TypeError):
        logger.info(f"Unable to open '{h5ad_path}' with AnnData")
        sys.exit(1)

    return adata


def get_hash_digest_column(dataframe):
    """
    Get column with hash digest for each row in dataframe.
    """
    df_index = dataframe.index.to_series()
    assert df_index.is_unique
    return (
        df_index.map(xxh3_64_intdigest)
        .astype(np.uint64)
        .apply(lambda v: b85encode(v.to_bytes(8, "big")).decode("ascii"))
    )


@lru_cache()
def is_ontological_descendant_of(onto: OntologyParser, term: str, target: str, include_self: bool = True) -> bool:
    """
    Determines if :term is an ontological descendant of :target and whether to include :term==:target.

    This function is cached and is safe to call many times.

    #TODO:[EM] needs testing
    """
    return term in set(onto.get_term_descendants(target, include_self))


@lru_cache()
def get_descendants(onto: OntologyParser, term: str, include_self: bool = True) -> List[str]:
    return onto.get_term_descendants(term, include_self=True)


def count_matrix_nonzero(matrix: DaskArray, is_sparse_matrix: bool) -> int:
    def count_nonzeros(matrix_chunk: Union[np.ndarray, sparse.spmatrix], is_sparse_matrix: bool) -> np.array:
        nnz = matrix_chunk.nnz if is_sparse_matrix else np.count_nonzero(matrix_chunk)
        return np.array([nnz])

    if len(matrix.chunks[0]) > 1:
        nonzeros = map_blocks(count_nonzeros, matrix, is_sparse_matrix, drop_axis=1, dtype=int).compute().sum()
    else:
        nonzeros = count_nonzeros(matrix.compute(), is_sparse_matrix)[0]
    return nonzeros


def check_non_csr_matrices(adata: ad.AnnData):
    """
    Check X, raw.X and layers matrices for having more than 50% zeros and not being csr_matrix

    If found, convert to csr_matrix
    """

    def get_sparsity(matrix: DaskArray, format: str):
        is_sparse_matrix = format in SPARSE_MATRIX_TYPES
        nnz = count_matrix_nonzero(matrix, is_sparse_matrix)
        sparsity = 1 - nnz / np.prod(matrix.shape)
        return sparsity

    format = get_matrix_format(adata.X)
    if format != "csr" and get_sparsity(adata.X, format) >= 0.5:
        adata.X = adata.X.map_blocks(sparse.csr_matrix, dtype=adata.X.dtype)

    if adata.raw is not None:
        format = get_matrix_format(adata.raw.X)
        if format != "csr" and get_sparsity(adata.raw.X, format) >= 0.5:
            raw_adata = ad.AnnData(adata.raw.X, var=adata.raw.var, obs=adata.obs)
            raw_adata.X = raw_adata.X.map_blocks(sparse.csr_matrix, dtype=raw_adata.X.dtype)
            adata.raw = raw_adata
            del raw_adata

    for layer in adata.layers:
        format = get_matrix_format(adata.layers[layer])
        if format != "csr" and get_sparsity(adata.layers[layer], format) >= 0.5:
            adata.layers[layer] = adata.layers[layer].map_blocks(sparse.csr_matrix, dtype=adata.layers[layer].X.dtype)

    return adata



import logging
from tempfile import TemporaryDirectory
from unittest import mock

import numpy
import pandas as pd
import pytest
from cellxgene_schema.map_species import OntologyParser, map_species
from cellxgene_schema.validate import Validator
from fixtures.examples_validate import adata


@pytest.fixture
def zebrafish_adata():
    zebrafish_adata = adata.copy()
    obs = zebrafish_adata.obs
    obs.drop(columns=["cell_type_ontology_term_id", "tissue_ontology_term_id"], inplace=True)
    obs["organism_ontology_term_id"] = "NCBITaxon:7955"
    obs["development_stage_ontology_term_id"] = "ZFS:0000016"
    obs["self_reported_ethnicity_ontology_term_id"] = "na"
    obs["tissue_type"] = "tissue"

    obs["organism_ontology_term_id"] = obs["organism_ontology_term_id"].astype("category")
    obs["development_stage_ontology_term_id"] = obs["development_stage_ontology_term_id"].astype("category")
    obs["self_reported_ethnicity_ontology_term_id"] = obs["self_reported_ethnicity_ontology_term_id"].astype("category")
    obs["tissue_type"] = obs["tissue_type"].astype("category")

    # single match
    obs["organism_cell_type_ontology_term_id"] = "ZFA:0005769"

    # single match
    obs["organism_tissue_ontology_term_id"] = "ZFA:0000047"

    zebrafish_adata.var = pd.DataFrame(
        [[False], [False], [False], [False], [False], [False], [False]],
        index=[
            "ENSDARG00000103202",
            "ENSDARG00000096156",
            "ENSDARG00000076160",
            "ENSDARG00000117163",
            "ENSDARG00000096187",
            "ENSDARG00000009657",
            "ENSDARG00000098510",
        ],
        columns=["feature_is_filtered"],
    )
    zebrafish_adata.X = numpy.ones([zebrafish_adata.obs.shape[0], zebrafish_adata.var.shape[0]], dtype=numpy.float32)
    zebrafish_adata.raw = zebrafish_adata.copy()
    zebrafish_adata.raw.var.drop("feature_is_filtered", axis=1, inplace=True)
    return zebrafish_adata


@pytest.fixture
def validator_with_adata(zebrafish_adata):
    validator = Validator()
    validator.adata = zebrafish_adata.copy()
    return validator


def test_map_species__valid_output(validator_with_adata):
    with TemporaryDirectory() as tmp:
        input_file = tmp + "input.h5ad"
        validator_with_adata.adata.copy().write_h5ad(input_file, compression="gzip")
        output_file = tmp + "output.h5ad"
        map_species(input_file, output_file)

        assert validator_with_adata.validate_adata(output_file) is True


def test_map_species_log_ouput(validator_with_adata, caplog):
    caplog.set_level(logging.INFO)
    obs = validator_with_adata.adata.obs
    # test single match, error match, multiple options
    obs.loc[obs.index[0], "organism_cell_type_ontology_term_id"] = "ZFA:0009384"
    obs.loc[obs.index[1], "organism_tissue_ontology_term_id"] = "ZFA:0"

    with TemporaryDirectory() as tmp:
        input_file = tmp + "input.h5ad"
        validator_with_adata.adata.copy().write_h5ad(input_file, compression="gzip")
        output_file = tmp + "output.h5ad"
        map_species(input_file, output_file)
        assert (
            "ZFA:0009384 has multiple closest matches for CL in its ancestry tree: ['CL:0000075', 'CL:0002077']"
            in caplog.text
        )
        assert "Setting ZFA:0005769 cell_type_ontology_term_id rows to single closest match: CL:0000018" in caplog.text
        assert (
            "ZFA:0 in organism_tissue_ontology_term_id is not a valid ontology term ID, cannot find UBERON map."
            in caplog.text
        )
        assert "Setting ZFA:0000047 tissue_ontology_term_id rows to single closest match: UBERON:0000004" in caplog.text

        # output should be invalid, per logs
        assert validator_with_adata.validate_adata(output_file) is False


def test_map_species_log_output_no_match(zebrafish_adata, caplog):
    with TemporaryDirectory() as tmp, mock.patch.object(
        OntologyParser, "get_closest_bridge_term_ids", mock.Mock(return_value=[])
    ):
        input_file = tmp + "input.h5ad"
        zebrafish_adata.copy().write_h5ad(input_file, compression="gzip")
        output_file = tmp + "output.h5ad"
        map_species(input_file, output_file)
        assert "ZFA:0005769 has no closest match for CL in its ancestry tree." in caplog.text
        assert "ZFA:0000047 has no closest match for UBERON in its ancestry tree." in caplog.text


def test_map_species__with_existing_cols(validator_with_adata):
    obs = validator_with_adata.adata.obs
    obs["cell_type_ontology_term_id"] = ""
    obs["tissue_ontology_term_id"] = ""
    # set-up organism_cell_type and organism_tissue term IDs with one of multiple matching CL/UBERON options
    # leave single match rows blank, to be set by map_species
    obs.loc[obs.index[0], "organism_cell_type_ontology_term_id"] = "ZFA:0009384"
    obs.loc[obs.index[0], "cell_type_ontology_term_id"] = "CL:0000075"
    obs.loc[obs.index[0], "organism_tissue_ontology_term_id"] = "ZFA:0005914"
    obs.loc[obs.index[0], "tissue_ontology_term_id"] = "UBERON:0001769"

    with TemporaryDirectory() as tmp:
        input_file = tmp + "input.h5ad"
        validator_with_adata.adata.copy().write_h5ad(input_file, compression="gzip")
        output_file = tmp + "output.h5ad"
        map_species(input_file, output_file)

        assert validator_with_adata.validate_adata(output_file) is True



import logging
import sys

import click

logger = logging.getLogger("cellxgene_schema")


@click.group(
    name="schema",
    subcommand_metavar="COMMAND <args>",
    short_help="Apply and validate the cellxgene data integration schema to an h5ad file.",
    context_settings=dict(max_content_width=85, help_option_names=["-h", "--help"]),
)
@click.option("-v", "--verbose", help="When present will set logging level to debug", is_flag=True)
def schema_cli(verbose):
    logging.basicConfig(level=logging.ERROR)
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)


@schema_cli.command(
    name="validate",
    short_help="Check that an h5ad follows the cellxgene data integration schema.",
    help="Check that an h5ad follows the cellxgene data integration schema. If validation fails this command will "
    "return an exit status of 1 otherwise 0. When the '--add-labels <FILE>' tag is present, the command will add "
    "ontology/gene labels based on IDs and write them to a new h5ad.",
)
@click.argument("h5ad_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.option(
    "-a",
    "--add-labels",
    "add_labels_file",
    help="When present it will add labels to genes and ontologies based on IDs",
    required=False,
    default=None,
    type=click.Path(exists=False, dir_okay=False, writable=True),
)
@click.option("-i", "--ignore-labels", help="Ignore ontology labels when validating", is_flag=True)
def schema_validate(h5ad_file, add_labels_file, ignore_labels):
    # Imports are very slow so we defer loading until Click arg validation has passed
    logger.info("Loading dependencies")
    try:
        import anndata  # noqa: F401
    except ImportError:
        raise click.ClickException("[cellxgene] cellxgene-schema requires anndata") from None

    logger.info("Loading validator modules")
    from .validate import validate

    is_valid, _, _ = validate(h5ad_file, add_labels_file, ignore_labels=ignore_labels)
    if is_valid:
        sys.exit(0)
    else:
        sys.exit(1)


@schema_cli.command(
    name="add-labels",
    short_help="Create a copy of an h5ad with portal-added labels",
    help="Create a copy of an h5ad with portal-added labels. The labels are added based on the IDs in the "
    "provided file.",
)
@click.argument("input_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("output_file", nargs=1, type=click.Path(exists=False, dir_okay=False))
def add_labels(input_file, output_file):
    from utils import read_h5ad

    from .write_labels import AnnDataLabelAppender

    logger.info(f"Loading h5ad from {input_file}")
    adata = read_h5ad(input_file)
    anndata_label_adder = AnnDataLabelAppender(adata)
    logger.info("Adding labels")
    if anndata_label_adder.write_labels(output_file):
        sys.exit(0)
    else:
        sys.exit(1)


def _check_anndata_requires_fragment(h5ad_file):
    from .atac_seq import check_anndata_requires_fragment, report_errors

    try:
        fragment_required = check_anndata_requires_fragment(h5ad_file)
        if fragment_required:
            logger.info("Anndata requires an ATAC fragment file.")
        else:
            logger.info("Anndata does not require an ATAC fragment file.")
    except Exception as e:
        report_errors("Anndata does not support ATAC fragment files for the followings reasons:", [str(e)])
        sys.exit(1)


@schema_cli.command(
    name="process-fragment",
    short_help="Check that an ATAC SEQ fragment follows the cellxgene data integration schema.",
    help="Check that an ATAC SEQ fragment follows the cellxgene data integration schema. If validation fails this "
    "command will return an exit status of 1 otherwise 0. When the '--generate-index' tag is present, "
    "the command will generate a tabix compatible version of the fragment and tabix index. The generated "
    "fragment will have the file suffix .bgz and the index will have the file suffix .bgz.tbi.",
)
@click.argument("h5ad_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("fragment_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.option("-i", "--generate-index", help="Generate index for fragment", is_flag=True)
@click.option("-o", "--output-file", help="Output file for the processed fragment.", type=click.Path(exists=False))
def fragment_validate(h5ad_file, fragment_file, generate_index, output_file):
    from .atac_seq import process_fragment

    _check_anndata_requires_fragment(h5ad_file)

    if not process_fragment(fragment_file, h5ad_file, generate_index=generate_index, output_file=output_file):
        sys.exit(1)


@schema_cli.command(
    name="check-anndata-requires-fragment",
    short_help="Check if that the anndata provided supports an Atac seq fragment file.",
    help="Check that an ATAC SEQ anndata.obs['assay_ontology_term_id'] is all paired or unpaired assays. "
    "This determines if a fragment file is required, optional, or forbiden. "
    "If the anndata does not support a fragment, an error message will be printed and exit status will be 1.",
)
@click.argument("h5ad_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
def check_anndata_requires_fragment(h5ad_file):
    _check_anndata_requires_fragment(h5ad_file)


@schema_cli.command(
    name="remove-labels",
    short_help="Create a copy of an h5ad without portal-added labels",
    help="Create a copy of an h5ad without portal-added labels.",
)
@click.argument("input_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("output_file", nargs=1, type=click.Path(exists=False, dir_okay=False))
def remove_labels(input_file, output_file):
    from .remove_labels import AnnDataLabelRemover

    logger.info("Loading dependencies")
    try:
        import anndata  # noqa: F401
    except ImportError:
        raise click.ClickException("[cellxgene] cellxgene-schema requires anndata") from None

    logger.info(f"Loading h5ad from {input_file}")
    adata = anndata.read_h5ad(input_file)
    anndata_label_remover = AnnDataLabelRemover(adata)
    if not anndata_label_remover.schema_def:
        return
    logger.info("Removing labels")
    anndata_label_remover.remove_labels()
    logger.info(f"Labels have been removed. Writing to {output_file}")
    anndata_label_remover.adata.write(output_file, compression="gzip")


@schema_cli.command(
    name="migrate",
    short_help="Convert an h5ad to the latest schema version.",
    help="Convert an h5ad from the previous to latest minor schema version. No validation will be "
    "performed on either the input or the output file.",
)
@click.argument("input_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("output_file", nargs=1, type=click.Path(exists=False, dir_okay=False))
@click.option("--collection_id", default=None, type=str, help="Collection ID, if migrating already published dataset")
@click.option("--dataset_id", default=None, type=str, help="Dataset ID, if migrating already published dataset")
def migrate(input_file, output_file, collection_id, dataset_id):
    from .migrate import migrate

    migrate(input_file, output_file, collection_id, dataset_id)


@schema_cli.command(
    name="map-species",
    short_help="Annotate non-human, non-mouse anndata with CL and UBERON equivalent terms",
    help="Annotate non-human, non-mouse anndata with CL and UBERON equivalent terms, based on values in"
    "organism-specific columns (e.g. organism_cell_type_ontology_term_id and organism_tissue_ontology_term_id)",
)
@click.argument("input_file", nargs=1, type=click.Path(exists=True, dir_okay=False))
@click.argument("output_file", nargs=1, type=click.Path(exists=False, dir_okay=False))
def map_species(input_file, output_file):
    from .map_species import map_species

    map_species(input_file, output_file)


if __name__ == "__main__":
    schema_cli()



import logging
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import Optional

import anndata as ad
import h5py
import ibis
import pyarrow as pa
import pyarrow.csv
import pyarrow.dataset
import pyarrow.parquet
import pysam

from .ontology_parser import ONTOLOGY_PARSER
from .utils import GB, is_ontological_descendant_of

logger = logging.getLogger(__name__)

# TODO: these chromosome tables should be calculated from the fasta file?
# location of fasta https://www.gencodegenes.org/human/release_44.html and file name GRCh38.primary_assembly.genome.fa
human_chromosome_by_length = {
    "chr1": 248956422,
    "chr2": 242193529,
    "chr3": 198295559,
    "chr4": 190214555,
    "chr5": 181538259,
    "chr6": 170805979,
    "chr7": 159345973,
    "chr8": 145138636,
    "chr9": 138394717,
    "chr10": 133797422,
    "chr11": 135086622,
    "chr12": 133275309,
    "chr13": 114364328,
    "chr14": 107043718,
    "chr15": 101991189,
    "chr16": 90338345,
    "chr17": 83257441,
    "chr18": 80373285,
    "chr19": 58617616,
    "chr20": 64444167,
    "chr21": 46709983,
    "chr22": 50818468,
    "chrX": 156040895,
    "chrY": 57227415,
    "chrM": 16569,
    "GL000009.2": 201709,
    "GL000194.1": 191469,
    "GL000195.1": 182896,
    "GL000205.2": 185591,
    "GL000213.1": 164239,
    "GL000216.2": 176608,
    "GL000218.1": 161147,
    "GL000219.1": 179198,
    "GL000220.1": 161802,
    "GL000225.1": 211173,
    "KI270442.1": 392061,
    "KI270711.1": 42210,
    "KI270713.1": 40745,
    "KI270721.1": 100316,
    "KI270726.1": 43739,
    "KI270727.1": 448248,
    "KI270728.1": 1872759,
    "KI270731.1": 150754,
    "KI270733.1": 179772,
    "KI270734.1": 165050,
    "KI270744.1": 168472,
    "KI270750.1": 148850,
}
mouse_chromosome_by_length = {
    "chr1": 195154279,
    "chr2": 181755017,
    "chr3": 159745316,
    "chr4": 156860686,
    "chr5": 151758149,
    "chr6": 149588044,
    "chr7": 144995196,
    "chr8": 130127694,
    "chr9": 124359700,
    "chr10": 130530862,
    "chr11": 121973369,
    "chr12": 120092757,
    "chr13": 120883175,
    "chr14": 125139656,
    "chr15": 104073951,
    "chr16": 98008968,
    "chr17": 95294699,
    "chr18": 90720763,
    "chr19": 61420004,
    "chrX": 169476592,
    "chrY": 91455967,
    "chrM": 16299,
    "GL456210.1": 169725,
    "GL456211.1": 241735,
    "GL456212.1": 153618,
    "GL456219.1": 175968,
    "GL456221.1": 206961,
    "GL456239.1": 40056,
    "GL456354.1": 195993,
    "GL456372.1": 28664,
    "GL456381.1": 25871,
    "GL456385.1": 35240,
    "JH584295.1": 1976,
    "JH584296.1": 199368,
    "JH584297.1": 205776,
    "JH584298.1": 184189,
    "JH584299.1": 953012,
    "JH584303.1": 158099,
    "JH584304.1": 114452,
}
organism_ontology_term_id_by_chromosome_length_table = {
    "NCBITaxon:9606": human_chromosome_by_length,
    "NCBITaxon:10090": mouse_chromosome_by_length,
}
column_ordering = ["chromosome", "start coordinate", "stop coordinate", "barcode", "read support"]
schema = pa.schema(
    [
        pa.field("chromosome", pa.string()),
        pa.field("start coordinate", pa.int64()),
        pa.field("stop coordinate", pa.int64()),
        pa.field("barcode", pa.string()),
        pa.field("read support", pa.int64()),
    ]
)


def log_calls(func):
    def wrapper(*args, **kwargs):
        logging.info("starting: %s", func.__name__)
        return func(*args, **kwargs)

    return wrapper


def is_atac(x: str) -> str:
    if is_ontological_descendant_of(ONTOLOGY_PARSER, x, "EFO:0010891"):
        if is_ontological_descendant_of(ONTOLOGY_PARSER, x, "EFO:0008913"):
            return "p"  # paired
        else:
            return "u"  # unpaired
    else:
        return "n"  # not atac seq


def check_anndata_requires_fragment(anndata_file: str) -> bool:
    """
    Check if an anndata file requires a fragment file to be valid. The anndata file requires a fragment file if the
    assay_ontology_term_id are all descendants of EFO:0010891 and EFO:0008913. If the assay_ontology_term_id are all
    descendants of EFO:0010891 and not EFO:0008913, the anndata file does not require a fragment file. If the
    assay_ontology_term_id are not all descendants of EFO:0010891, an error is raised and a fragment file is not allowed
    because the anndata is not a valid ATAC-Seq file.

    :param anndata_file: The anndata file to validate.
    :return:
    """
    with h5py.File(anndata_file) as f:
        assay_ontology_term_ids = ad.io.read_elem(f["obs"])["assay_ontology_term_id"]
    df = assay_ontology_term_ids.map(is_atac)

    if (df == "p").all():
        return False
    elif (df == "u").all():
        return True
    if (df == "n").any():
        raise ValueError("Anndata.obs.assay_ontology_term_id are not all descendants of EFO:0010891.")
    else:
        raise ValueError("Anndata.obs.assay_ontology_term_id has mixed paired and unpaired assay terms.")


def process_fragment(
    fragment_file: str,
    anndata_file: str,
    generate_index: bool = False,
    override_write_algorithm: Optional[str] = None,
    output_file: Optional[str] = None,
) -> list[str]:
    """
    Validate the fragment against the anndata file and generate the index if the fragment is valid.

    :param str fragment_file: The fragment file to process
    :param str anndata_file: The anndata file to validate against
    :param bool generate_index: Whether to generate the index for the fragment
    :param override_write_algorithm: Override the write algorithm used to write the bgzip file. Options are "pysam"
    and "cli"
    :param output_file: The output file to write the bgzip file to. If not provided, the output file will be the same

    """
    with tempfile.TemporaryDirectory() as tempdir:
        # quick checks
        errors = validate_anndata(anndata_file)
        if errors:
            return errors

        # convert the fragment to a parquet file for faster processing
        try:
            parquet_file = convert_to_parquet(fragment_file, tempdir)
        except Exception as e:
            msg = "Error Parsing the fragment file. Check that columns match schema definition. Error: " + str(e)
            logger.exception(msg)
            return [msg]

        # slow checks
        errors = validate_anndata_with_fragment(parquet_file, anndata_file)
        if errors:
            return errors
        else:
            logger.info("Fragment and Anndata file are valid")

        # generate the index
        if generate_index:
            logger.info(f"Sorting fragment and generating index for {fragment_file}")
            index_fragment(fragment_file, parquet_file, tempdir, override_write_algorithm, output_file)
    logger.debug("cleaning up")
    return []


def convert_to_parquet(fragment_file: str, tempdir: str) -> str:
    """
    Convert the fragment file to a parquet dataset for faster processing.

    :param fragment_file: A gzipped compressed fragment file
    :param tempdir: The temporary directory to write the parquet file to. Name of the written file is derived from
    the input.
    """
    logger.info(f"Converting {fragment_file} to parquet")
    parquet_file_path = Path(tempdir) / Path(fragment_file).name.replace(".gz", ".parquet").replace(".bgz", ".parquet")
    pa.dataset.write_dataset(
        data=pa.csv.open_csv(
            pa.input_stream(fragment_file, compression="gzip", buffer_size=GB),
            read_options=pa.csv.ReadOptions(column_names=schema.names),
            parse_options=pa.csv.ParseOptions(delimiter="\t"),
            convert_options=pa.csv.ConvertOptions(column_types=schema),
        ),
        base_dir=parquet_file_path,
        format="parquet",
        # Using hive partitioning for best compatibility with dask to_parquet and read_parquet functions
        partitioning=pa.dataset.partitioning(pa.schema([pa.field("chromosome", pa.string())]), flavor="hive"),
    )

    return str(parquet_file_path)


def report_errors(header: str, errors: list[str]) -> list[str]:
    if any(errors):
        errors = [e for e in errors if e is not None]
        errors = [header] + errors
        logger.error("\n\t".join(errors))
        return errors
    else:
        return []


def validate_anndata(anndata_file: str) -> list[str]:
    errors = [validate_anndata_organism_ontology_term_id(anndata_file), validate_anndata_is_primary_data(anndata_file)]
    return report_errors("Errors found in Anndata file. Skipping fragment validation.", errors)


def validate_anndata_with_fragment(parquet_file: str, anndata_file: str) -> list[str]:
    errors = [
        validate_fragment_start_coordinate_greater_than_0(parquet_file),
        validate_fragment_barcode_in_adata_index(parquet_file, anndata_file),
        validate_fragment_stop_coordinate_within_chromosome(parquet_file, anndata_file),
        validate_fragment_stop_greater_than_start_coordinate(parquet_file),
        validate_fragment_read_support(parquet_file),
        validate_fragment_no_duplicate_rows(parquet_file),
    ]
    return report_errors("Errors found in Fragment and/or Anndata file", errors)


@log_calls
def validate_fragment_no_duplicate_rows(parquet_file: str) -> Optional[str]:
    t = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    rows_per_chromosome = t["chromosome"].value_counts().execute()
    msg = ""
    # Checking number of unique rows per chromosome is more memory efficient than checking all rows at once
    for chromosome, count in rows_per_chromosome.itertuples(index=False):
        n_unique = t.filter(t["chromosome"] == chromosome).distinct().count().execute()
        if n_unique != count:
            if not msg:
                msg = "Fragment file has duplicate rows.\n"
            msg += f"Chromosome {chromosome} has {count} rows but only {n_unique} are unique\n"
    if msg:
        return msg.strip()  # remove trailing newline


@log_calls
def validate_fragment_start_coordinate_greater_than_0(parquet_file: str) -> Optional[str]:
    df = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    if not (df["start coordinate"] > 0).all().execute():
        return "Start coordinate must be greater than 0."


@log_calls
def validate_fragment_barcode_in_adata_index(parquet_file: str, anndata_file: str) -> Optional[str]:
    df = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    barcode = set(df.select("barcode").distinct().execute()["barcode"])
    with h5py.File(anndata_file) as f:
        obs = ad.io.read_elem(f["obs"])
    if set(obs.index) != barcode:
        return "Barcodes don't match anndata.obs.index"


@log_calls
def validate_fragment_stop_greater_than_start_coordinate(parquet_file: str) -> Optional[str]:
    df = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    if not (df["stop coordinate"] > df["start coordinate"]).all().execute():
        return "Stop coordinate must be greater than start coordinate."


@log_calls
def validate_fragment_stop_coordinate_within_chromosome(parquet_file: str, anndata_file: str) -> Optional[str]:
    with h5py.File(anndata_file) as f:
        organism_ontology_term_id = ad.io.read_elem(f["obs"])["organism_ontology_term_id"].unique().astype(str)[0]
    chromosome_length_table = organism_ontology_term_id_by_chromosome_length_table[organism_ontology_term_id]
    t = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    df = t.group_by("chromosome").aggregate(max_stop_coordinate=t["stop coordinate"].max()).execute()

    mismatched_chromosomes = set(df["chromosome"].unique()) - chromosome_length_table.keys()
    if mismatched_chromosomes:
        return f"Chromosomes in the fragment do not match the organism({organism_ontology_term_id}).\n" + "\t\n".join(
            mismatched_chromosomes
        )

    df["chromosome_length"] = df["chromosome"].map(chromosome_length_table)
    if not (df["max_stop_coordinate"] <= df["chromosome_length"]).all():
        return "Stop coordinate must be less than the chromosome length."


@log_calls
def validate_fragment_read_support(parquet_file: str) -> Optional[str]:
    t = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    if (t["read support"] <= 0).any().execute():
        return "Read support must be greater than 0."


@log_calls
def validate_anndata_is_primary_data(anndata_file: str) -> Optional[str]:
    with h5py.File(anndata_file) as f:
        is_primary_data = ad.io.read_elem(f["obs"])["is_primary_data"]
    if not is_primary_data.all():
        return "Anndata.obs.is_primary_data must all be True."


@log_calls
def validate_anndata_organism_ontology_term_id(anndata_file: str) -> Optional[str]:
    with h5py.File(anndata_file) as f:
        organism_ontology_term_ids = ad.io.read_elem(f["obs"])["organism_ontology_term_id"].unique().astype(str)
    if organism_ontology_term_ids.size > 1:
        error_message = (
            "Anndata.obs.organism_ontology_term_id must have exactly 1 unique value. Found the " "following values:\n"
        ) + "\n\t".join(organism_ontology_term_ids)
        return error_message
    organism_ontology_term_id = organism_ontology_term_ids[0]
    allowed_terms = [*organism_ontology_term_id_by_chromosome_length_table.keys()]
    if organism_ontology_term_id not in allowed_terms:
        return f"Anndata.obs.organism_ontology_term_id must be one of {allowed_terms}. Got {organism_ontology_term_id}."


@log_calls
def detect_chromosomes(parquet_file: str) -> list[str]:
    t = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    chromosomes = list(t.select(["chromosome"]).distinct().execute()["chromosome"])
    chromosomes.sort()  # sort chromosomes to ensure consistent order
    return chromosomes


def get_output_file(fragment_file: str, output_file: Optional[str]) -> str:
    if not output_file:
        bgzip_output_file = fragment_file.replace(".gz", ".bgz")
    elif not output_file.endswith(".bgz"):
        bgzip_output_file = output_file + ".bgz"
    else:
        bgzip_output_file = output_file
    return bgzip_output_file


def index_fragment(
    fragment_file: str,
    parquet_file: str,
    tempdir: str,
    override_write_algorithm: Optional[str] = None,
    output_file: Optional[str] = None,
):
    # sort the fragment by chromosome, start coordinate, and stop coordinate, then compress it with bgzip
    bgzip_output_file = get_output_file(fragment_file, output_file)
    bgzip_output_path = Path(bgzip_output_file)
    bgzip_output_path.unlink(missing_ok=True)
    bgzip_output_path.touch()

    if override_write_algorithm:
        write_algorithm = write_algorithm_by_callable[override_write_algorithm]
    elif not shutil.which("bgzip"):  # check if bgzip cli is installed
        logger.warning("bgzip is not installed, using slower pysam implementation")
        write_algorithm = write_algorithm_by_callable["pysam"]
    else:
        write_algorithm = write_algorithm_by_callable["cli"]

    chromosomes = detect_chromosomes(parquet_file)
    prepare_fragment(chromosomes, parquet_file, bgzip_output_file, tempdir, write_algorithm)
    logger.info(f"Fragment sorted and compressed: {bgzip_output_file}")
    #
    pysam.tabix_index(bgzip_output_file, preset="bed", force=True)
    tabix_output_file = bgzip_output_file + ".tbi"
    logger.info(f"Index file generated: {tabix_output_file}")


def sort_fragment(parquet_file: str, write_path: str, chromosome: str) -> Path:
    temp_data = Path(write_path) / f"temp_{chromosome}.parquet"
    t = ibis.read_parquet(f"{parquet_file}/**", hive_partitioning=True)
    (t.filter(t["chromosome"] == chromosome).order_by(["start coordinate", "stop coordinate"]).to_parquet(temp_data))
    return temp_data


def buffered_write(input_file: str) -> iter:
    # Open the Parquet file and iterate through record batches
    pfile = pa.parquet.ParquetFile(input_file)
    for record_batch in pfile.iter_batches():
        # Write the batch to an in-memory buffer
        csv_buffer = pa.BufferOutputStream()
        pa.csv.write_csv(
            # Make sure columns are in right order
            record_batch.select([f.name for f in schema]),
            csv_buffer,
            write_options=pa.csv.WriteOptions(
                include_header=False,
                delimiter="\t",
                batch_size=1_000_000_000,  # this value could be further optimized
                quoting_style="none",
            ),
        )
        yield csv_buffer.getvalue().to_pybytes()


def write_bgzip_pysam(input_file: str, bgzip_output_file: str):
    with pysam.libcbgzf.BGZFile(bgzip_output_file, mode="ab") as f_out:
        for data in buffered_write(input_file):
            f_out.write(data)


def write_bgzip_cli(input_file: str, bgzip_output_file: str):
    with (
        subprocess.Popen(
            ["bgzip", "--threads", "-c"], stdin=subprocess.PIPE, stdout=open(bgzip_output_file, "ab")
        ) as proc,
    ):
        for data in buffered_write(input_file):
            proc.stdin.write(data)
    return_code = proc.wait()
    if return_code != 0:
        logger.error(f"Subprocess exited with error code {return_code}")


write_algorithm_by_callable = {"pysam": write_bgzip_pysam, "cli": write_bgzip_cli}


def prepare_fragment(
    chromosomes: list[str],
    parquet_file: str,
    bgzip_output_file: str,
    tempdir: str,
    write_algorithm: callable,
) -> None:
    """
    The sorting and writing of the fragment is done for each chromosome. Because of this the write order of
    the chromosomes may not be preserved. The chromosomes will all be stored in contiguous blocks in the bgzip file, and
    and sorted by start and stop coordinate within each chromosome.

    :param chromosomes:
    :param parquet_file:
    :param bgzip_output_file:
    :param tempdir:
    :param write_algorithm:
    :return:
    """
    for chromosome in chromosomes:
        logger.info(f"Processing chromosome: {chromosome}")
        temp_data = sort_fragment(parquet_file, tempdir, chromosome)
        write_algorithm(temp_data, bgzip_output_file)
    logger.info(f"bgzip compression completed successfully for {bgzip_output_file}")



import os
from pathlib import Path

import anndata as ad
import dask.dataframe as dd
import pandas as pd
import pysam
import pytest
from cellxgene_schema import atac_seq
from fixtures.examples_validate import FIXTURES_ROOT


@pytest.fixture
def atac_fragment_bgzip_file_path(tmpdir) -> Path:
    bgzip_file = Path(tmpdir + "new.tsv.bgz")
    yield bgzip_file
    bgzip_file.unlink(missing_ok=True)


@pytest.fixture
def atac_fragment_index_file_path(atac_fragment_bgzip_file_path) -> Path:
    index_file = Path(str(atac_fragment_bgzip_file_path) + ".tbi")
    yield index_file
    index_file.unlink(missing_ok=True)


def to_anndata_file(adata: ad.AnnData, path: str) -> str:
    file_name = os.path.join(path, "small_atac_seq.h5ad")
    adata.write(file_name)
    return file_name


@pytest.fixture
def atac_anndata_file(atac_anndata, tmpdir):
    file_name = os.path.join(tmpdir, "small_atac_seq.h5ad")
    atac_anndata.write(file_name)
    return file_name


def to_parquet_file(df: pd.DataFrame, path: str) -> str:
    file_name = os.path.join(path, "fragment")
    dd.from_pandas(df).to_parquet(file_name, partition_on=["chromosome"])
    return file_name


@pytest.fixture
def atac_fragment_dataframe() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "barcode": ["A", "B", "C"],
            "start coordinate": [100, 200, 300],
            "stop coordinate": [200, 300, 400],
            "read support": [1, 2, 3],
            "chromosome": ["chr1", "chr2", "chr3"],
        }
    )


@pytest.fixture
def atac_fragment_file(atac_fragment_dataframe, tmpdir):
    return to_parquet_file(atac_fragment_dataframe, tmpdir)


def count_fragments_per_chromosome(fragment_file):
    fragments = pd.read_csv(
        fragment_file,
        compression="gzip",
        sep="\t",
        header=None,
        names=["chromosome", "start coordinate", "stop coordinate", "barcode", "read support"],
    )
    fragments_per_chromosome = fragments["chromosome"].value_counts().sort_index()
    return fragments_per_chromosome


class TestProcessFragment:
    @pytest.mark.parametrize("override_write_algorithm", ["pysam", "cli", None])
    def test_write_algorithms(
        self, atac_fragment_bgzip_file_path, atac_fragment_index_file_path, override_write_algorithm
    ):
        self._test_process_fragment(
            atac_fragment_bgzip_file_path, atac_fragment_index_file_path, override_write_algorithm, "fragments.tsv.bgz"
        )

    @pytest.mark.parametrize("fragment_file", ["fragments.tsv.bgz", "fragments.tsv.gz"])
    def test_source_file_compression(self, atac_fragment_bgzip_file_path, atac_fragment_index_file_path, fragment_file):
        self._test_process_fragment(
            atac_fragment_bgzip_file_path, atac_fragment_index_file_path, "pysam", fragment_file
        )

    @staticmethod
    def _test_process_fragment(
        atac_fragment_bgzip_file_path, atac_fragment_index_file_path, override_write_algorithm, fragment_file
    ):
        # Arrange
        anndata_file = os.path.join(FIXTURES_ROOT, "atac_seq", "small_atac_seq.h5ad")
        fragments_file = os.path.join(FIXTURES_ROOT, "atac_seq", fragment_file)
        # Act
        result = atac_seq.process_fragment(
            fragments_file,
            anndata_file,
            generate_index=True,
            override_write_algorithm=override_write_algorithm,
            output_file=str(atac_fragment_bgzip_file_path),
        )
        # Assert
        assert len(result) == 0
        assert atac_fragment_bgzip_file_path.exists()
        # Testing the bgzip file
        with pysam.libcbgzf.BGZFile(atac_fragment_bgzip_file_path, mode="r") as bgzip:
            previous_stop, previous_start, previous_chomosome = 0, 0, None
            for line in bgzip:
                chromosome, start, stop, _, _ = line.decode("utf-8").split("\t")
                if previous_chomosome != chromosome:
                    previous_stop, previous_start = 0, 0
                    chromosome_length = atac_seq.human_chromosome_by_length[chromosome]
                start, stop = int(start), int(stop)
                assert start <= stop < chromosome_length, (
                    "Stop coordinate must be within the chromosome length and " "greater than start."
                )
                assert start >= 0, "Start coordinate must be greater than 0."
                # Fragment is sorted by start, then stop in ascending order
                assert start >= previous_start
                # The bellow check is failing.
                if (chromosome == previous_chomosome) and (start == previous_start):
                    assert stop >= previous_stop
                previous_stop = stop
                previous_start = start
                previous_chomosome = chromosome

        # check that the number of fragments per chromosome is the same
        orig_chrom_counts = count_fragments_per_chromosome(fragments_file)
        new_chrom_counts = count_fragments_per_chromosome(atac_fragment_bgzip_file_path)
        pd.testing.assert_series_equal(orig_chrom_counts, new_chrom_counts)

        # Testing index access
        assert atac_fragment_index_file_path.exists()
        with pysam.TabixFile(str(atac_fragment_bgzip_file_path)) as tabix:
            for chromosome in tabix.contigs:
                assert chromosome in atac_seq.human_chromosome_by_length
                assert orig_chrom_counts[chromosome] == len(list(tabix.fetch(chromosome)))

    def test_fail(self, atac_fragment_bgzip_file_path, atac_fragment_index_file_path):
        # Arrange
        anndata_file = os.path.join(FIXTURES_ROOT, "atac_seq", "small_atac_seq.h5ad")
        fragments_file = os.path.join(FIXTURES_ROOT, "atac_seq", "fragments_bad.tsv.gz")
        # Act
        result = atac_seq.process_fragment(
            fragments_file,
            anndata_file,
            generate_index=True,
            output_file=str(atac_fragment_bgzip_file_path),
        )
        result = [r for r in result if "Error" in r]


class TestConvertToParquet:
    def test_positive(self, atac_fragment_dataframe, tmpdir):
        tsv_file = os.path.join(
            tmpdir,
            "fragment.tsv.gz",
        )
        atac_fragment_dataframe.to_csv(
            tsv_file, sep="\t", index=False, compression="gzip", header=False, columns=atac_seq.column_ordering
        )
        parquet_file = Path(atac_seq.convert_to_parquet(tsv_file, tmpdir))
        assert Path(parquet_file).is_dir()

    def test_missing_column(self, tmpdir, atac_fragment_dataframe):
        atac_fragment_dataframe = atac_fragment_dataframe.drop(columns=["read support"])
        tsv_file = os.path.join(tmpdir, "fragment.tsv")
        atac_fragment_dataframe.to_csv(
            tsv_file, sep="\t", index=False, compression="gzip", header=False, columns=atac_seq.column_ordering[:-1]
        )
        with pytest.raises(ValueError):
            atac_seq.convert_to_parquet(tsv_file, tmpdir)

    def test_invalid_column_dtype(self, tmpdir, atac_fragment_dataframe):
        atac_fragment_dataframe["start coordinate"] = "foo"
        tsv_file = os.path.join(tmpdir, "fragment.tsv.gz")
        atac_fragment_dataframe.to_csv(
            tsv_file, sep="\t", index=False, compression="gzip", header=False, columns=atac_seq.column_ordering
        )
        with pytest.raises(ValueError):
            atac_seq.convert_to_parquet(tsv_file, tmpdir)

    def test_with_na_columns(self, atac_fragment_dataframe, tmpdir):
        atac_fragment_dataframe["barcode"] = pd.NA
        tsv_file = os.path.join(tmpdir, "fragment.tsv.gz")
        atac_fragment_dataframe.to_csv(
            tsv_file, sep="\t", index=False, compression="gzip", header=False, columns=atac_seq.column_ordering
        )
        parquet_file = Path(atac_seq.convert_to_parquet(tsv_file, tmpdir))
        parquet_df = dd.read_parquet(parquet_file, columns=["barcode"])
        assert len(parquet_df[parquet_df["barcode"] != ""].compute()) == 0


class TestValidateFragmentBarcodeInAdataIndex:
    def test_postive(self, atac_fragment_file, atac_anndata_file):
        result = atac_seq.validate_fragment_barcode_in_adata_index(atac_fragment_file, atac_anndata_file)
        assert not result

    def test_missmatch_anndata(self, atac_fragment_file, tmpdir):
        # Arrange
        atac_anndata_file = to_anndata_file(
            ad.AnnData(obs=pd.DataFrame(index=["A", "B", "E"]), var=pd.DataFrame()), tmpdir
        )
        # Act
        result = atac_seq.validate_fragment_barcode_in_adata_index(atac_fragment_file, atac_anndata_file)
        # Assert
        assert result == "Barcodes don't match anndata.obs.index"

    def test_missing_in_anndata(self, atac_fragment_file, atac_anndata_file, tmpdir):
        # Arrange
        atac_anndata_file = to_anndata_file(ad.AnnData(obs=pd.DataFrame(index=["A", "B"]), var=pd.DataFrame()), tmpdir)
        # Act
        result = atac_seq.validate_fragment_barcode_in_adata_index(atac_fragment_file, atac_anndata_file)
        # Assert
        assert result == "Barcodes don't match anndata.obs.index"

    def test_missmatch_in_parquet(self, atac_fragment_dataframe, atac_anndata_file, tmpdir):
        # Arrange
        atac_fragment_dataframe["barcode"] = ["A", "B", "E"]
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_barcode_in_adata_index(fragment_file, atac_anndata_file)
        # Assert
        assert result == "Barcodes don't match anndata.obs.index"

    def test_missing_in_parquet(self, atac_fragment_dataframe, atac_anndata_file, tmpdir):
        # Arrange
        fragment_dataframe = atac_fragment_dataframe.drop(index=2)
        fragment_file = to_parquet_file(fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_barcode_in_adata_index(fragment_file, atac_anndata_file)
        # Assert
        assert result == "Barcodes don't match anndata.obs.index"


class TestValidateFragmentStartCoordianteGreaterThan0:
    def test_positive(self, atac_fragment_file):
        result = atac_seq.validate_fragment_start_coordinate_greater_than_0(atac_fragment_file)
        assert not result

    def test_negative(self, atac_fragment_dataframe, tmpdir):
        # Arrange
        atac_fragment_dataframe["start coordinate"] = -1
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_start_coordinate_greater_than_0(fragment_file)
        # Assert
        assert result == "Start coordinate must be greater than 0."


class TestValidateFragmentStopCoordinateGreaterThanStartCoordinate:
    def test_positive(self, atac_fragment_file):
        result = atac_seq.validate_fragment_stop_greater_than_start_coordinate(atac_fragment_file)
        assert not result

    def test_negative(self, atac_fragment_dataframe, tmpdir):
        # Arrange
        atac_fragment_dataframe["stop coordinate"] = 1
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_stop_greater_than_start_coordinate(fragment_file)
        # Assert
        assert result == "Stop coordinate must be greater than start coordinate."


class TestValidateFragmentStopCoordinateWithinChromosome:
    def test_positive(self, atac_fragment_file, atac_anndata_file):
        result = atac_seq.validate_fragment_stop_coordinate_within_chromosome(atac_fragment_file, atac_anndata_file)
        assert not result

    def test_stop_less_than_chromosome_length(self, atac_fragment_dataframe, atac_anndata_file, tmpdir):
        # Arrange
        atac_fragment_dataframe["stop coordinate"] = 10e12
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_stop_coordinate_within_chromosome(fragment_file, atac_anndata_file)
        # Assert
        assert result == "Stop coordinate must be less than the chromosome length."

    def test_mismatch_chromosome(self, atac_fragment_dataframe, atac_anndata_file, tmpdir):
        # Arrange
        atac_fragment_dataframe["chromosome"] = ["foo", "chr2", "chr1"]
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_stop_coordinate_within_chromosome(fragment_file, atac_anndata_file)
        # Assert
        assert result.startswith("Chromosomes in the fragment do not match the organism")


class TestValidateFragmentReadSupport:
    def test_positive(self, atac_fragment_file):
        result = atac_seq.validate_fragment_read_support(atac_fragment_file)
        assert not result

    @pytest.mark.parametrize("read_support", [0, -1])
    def test_negative(self, atac_fragment_dataframe, tmpdir, read_support):
        # Arrange
        atac_fragment_dataframe["read support"] = read_support
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_read_support(fragment_file)
        # Assert
        assert result == "Read support must be greater than 0."


class TestCheckAnndataRequiresFragment:

    @pytest.mark.parametrize("assay_ontology_term_id,expected_result", [("EFO:0030059", False), ("EFO:0030007", True)])
    def test_positive(self, atac_anndata, assay_ontology_term_id, expected_result, tmpdir):
        # Arrange
        atac_anndata.obs["assay_ontology_term_id"] = assay_ontology_term_id
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        # Act
        result = atac_seq.check_anndata_requires_fragment(atac_anndata_file)
        # Assert
        assert result == expected_result

    def test_not_atac(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["assay_ontology_term_id"] = "EFO:0030060"
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        with pytest.raises(
            ValueError, match="Anndata.obs.assay_ontology_term_id are not all descendants of EFO:0010891."  # Assert
        ):
            # Act
            atac_seq.check_anndata_requires_fragment(atac_anndata_file)

    def test_mixed_paired_and_unpaired(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["assay_ontology_term_id"] = ["EFO:0030059", "EFO:0030007", "EFO:0030007"]
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        with pytest.raises(
            ValueError, match="Anndata.obs.assay_ontology_term_id has mixed paired and unpaired assay terms."
        ):  # Assert
            # Act
            atac_seq.check_anndata_requires_fragment(atac_anndata_file)


class TestValidateAnndataIsPrimaryData:
    def test_positive(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["is_primary_data"] = True
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        # Act
        result = atac_seq.validate_anndata_is_primary_data(atac_anndata_file)
        # Assert
        assert not result

    def test_negative(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["is_primary_data"] = False
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        # Act
        result = atac_seq.validate_anndata_is_primary_data(atac_anndata_file)
        # Assert
        assert result == "Anndata.obs.is_primary_data must all be True."


class TestValidateAnndataOrganismOntologyTermId:
    def test_positive(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["organism_ontology_term_id"] = ["NCBITaxon:9606"] * 3
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        # Act
        result = atac_seq.validate_anndata_organism_ontology_term_id(atac_anndata_file)
        # Assert
        assert not result

    def test_organism_ontology_term_id_not_allowed(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["organism_ontology_term_id"] = ["NCBITaxon:9607"] * 3
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        # Act
        result = atac_seq.validate_anndata_organism_ontology_term_id(atac_anndata_file)
        # Assert
        assert result.startswith(
            "Anndata.obs.organism_ontology_term_id must be one of ['NCBITaxon:9606', 'NCBITaxon:10090']."
        )

    def test_organism_ontology_id_not_unique(self, atac_anndata, tmpdir):
        # Arrange
        atac_anndata.obs["organism_ontology_term_id"] = ["NCBITaxon:10090", "NCBITaxon:9606", "NCBITaxon:9606"]
        atac_anndata_file = to_anndata_file(atac_anndata, tmpdir)
        # Act
        result = atac_seq.validate_anndata_organism_ontology_term_id(atac_anndata_file)
        # Assert
        assert result.startswith("Anndata.obs.organism_ontology_term_id must have exactly 1 unique value.")


class TestValidateFragmentNoDuplicateRows:
    def test_positive(self, atac_fragment_file):
        result = atac_seq.validate_fragment_no_duplicate_rows(atac_fragment_file)
        assert not result

    def test_negative(self, atac_fragment_dataframe, tmpdir):
        # Arrange
        atac_fragment_dataframe = pd.concat([atac_fragment_dataframe, atac_fragment_dataframe])
        fragment_file = to_parquet_file(atac_fragment_dataframe, tmpdir)
        # Act
        result = atac_seq.validate_fragment_no_duplicate_rows(fragment_file)
        # Assert
        assert result.startswith("Fragment file has duplicate rows.")
        for chrom in atac_fragment_dataframe["chromosome"].unique():
            assert f"Chromosome {chrom} has 2 rows but only 1 are unique" in result


class TestGetOutputFile:
    @pytest.mark.parametrize(
        "fragment_file,output_file,expected",
        [
            ("fragment.gz", None, "fragment.bgz"),
            ("fragment.gz", "output.bgz", "output.bgz"),
            ("fragment.gz", "output", "output.bgz"),
        ],
    )
    def test_none(self, fragment_file, output_file, expected):
        assert atac_seq.get_output_file(fragment_file, output_file) == expected
